# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.training.checkpoints."""

import copy
import os
import pathlib
from typing import Any

import jax
import numpy as np
import orbax.checkpoint as orbax
from absl.testing import absltest, parameterized
from jax import numpy as jnp

from flax import config, core, errors, io, struct
from flax import linen as nn
from flax.training import checkpoints

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


PyTree = Any


def check_eq(xs, ys):
  return jax.tree_util.tree_all(
    jax.tree_util.tree_map(np.testing.assert_allclose, xs, ys)
  )


def shuffle(l):
  """Functional shuffle."""
  l = copy.copy(l)
  np.random.shuffle(l)
  return l


class Inner(nn.Module):
  """Inner class based on nn."""

  @nn.compact
  def __call__(self, x):
    x = nn.Conv(10, (2, 2))(x)
    x = nn.normalization.BatchNorm(True)(x)
    return x


class Model(nn.Module):
  """Simple model based on nn."""

  @nn.compact
  def __call__(self, inputs):
    x = nn.Conv(10, (2, 2))(inputs)
    x = Inner()(x)
    x = x.reshape([x.shape[0], -1])
    x = nn.normalization.BatchNorm(True)(x)
    x = nn.Dense(10)(x)
    x = nn.log_softmax(x)
    return x


@struct.dataclass
class CustomDC:
  foo: Any
  bar: Any


class CheckpointsTest(parameterized.TestCase):
  def setUp(self):
    super().setUp()
    config.update('flax_use_orbax_checkpointing', False)  # default value

  def test_naturalsort(self):
    np.random.seed(0)
    tests = [
      ['file_1', 'file_2', 'file_10', 'file_11', 'file_21'],
      ['file_0.001', 'file_0.01', 'file_0.1', 'file_1'],
      ['file_-3.0', 'file_-2', 'file_-1', 'file_0.0'],
      ['file_1e1', 'file_1.0e2', 'file_1e3', 'file_1.0e4'],
      ['file_1', 'file_2', 'file_9', 'file_1.0e1', 'file_11'],
    ]
    for test in tests:
      self.assertEqual(test, checkpoints.natural_sort(shuffle(test)))

  def test_safe_normpath(self):
    tests = ['./a/b/c', '/a//b/c', '/a/../b/c', 'a/b/./c', 'gs://a//b/c']
    expected = ['a/b/c', '/a/b/c', '/b/c', 'a/b/c', 'gs://a/b/c']
    for test, expect in zip(tests, expected):
      self.assertEqual(expect, checkpoints.safe_normpath(test))

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_save_restore_checkpoints(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    test_object0 = {
      'a': np.array([0, 0, 0], np.int32),
      'b': np.array([0, 0, 0], np.int32),
    }
    test_object1 = {
      'a': np.array([1, 2, 3], np.int32),
      'b': np.array([1, 1, 1], np.int32),
    }
    test_object2 = {
      'a': np.array([4, 5, 6], np.int32),
      'b': np.array([2, 2, 2], np.int32),
    }
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, prefix='test_'
    )
    check_eq(new_object, test_object0)
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 0, prefix='test_', keep=1
    )
    self.assertIn('test_0', os.listdir(tmp_dir))
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, prefix='test_'
    )
    check_eq(new_object, test_object1)
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 1, prefix='test_', keep=1
    )
    checkpoints.save_checkpoint(
      tmp_dir, test_object2, 2, prefix='test_', keep=1
    )
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, prefix='test_'
    )
    check_eq(new_object, test_object2)
    checkpoints.save_checkpoint(
      tmp_dir, test_object2, 3, prefix='test_', keep=2
    )
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 4, prefix='test_', keep=2
    )
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, prefix='test_'
    )
    check_eq(new_object, test_object1)
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, step=3, prefix='test_'
    )
    check_eq(new_object, test_object2)

    # Restore with a specific checkpoint path, not the directory path.
    new_object = checkpoints.restore_checkpoint(
      os.path.join(tmp_dir, 'test_3'), test_object0
    )
    check_eq(new_object, test_object2)
    # If a specific path is specified, but it does not exist, the same behavior
    # as when a directory is empty should apply: the target is returned
    # unchanged.
    new_object = checkpoints.restore_checkpoint(
      os.path.join(tmp_dir, 'test_not_there'), test_object0
    )
    check_eq(new_object, test_object0)
    with self.assertRaises(ValueError):
      checkpoints.restore_checkpoint(
        tmp_dir, test_object0, step=5, prefix='test_'
      )

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_overwrite_checkpoints(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    overwrite_error = ValueError if use_orbax else errors.InvalidCheckpointError
    tmp_dir = self.create_tempdir().full_path
    test_object0 = {'a': np.array([0, 0, 0], np.int32)}
    test_object = {'a': np.array([1, 2, 3], np.int32)}

    checkpoints.save_checkpoint(tmp_dir, test_object0, 0, keep=1)
    with self.assertRaises(overwrite_error):
      checkpoints.save_checkpoint(tmp_dir, test_object, 0, keep=1)
    checkpoints.save_checkpoint(tmp_dir, test_object, 0, keep=1, overwrite=True)
    new_object = checkpoints.restore_checkpoint(tmp_dir, test_object0)
    check_eq(new_object, test_object)

    non_norm_dir_path = tmp_dir + '//'
    checkpoints.save_checkpoint(non_norm_dir_path, test_object, 4, keep=1)
    new_object = checkpoints.restore_checkpoint(non_norm_dir_path, test_object0)
    check_eq(new_object, test_object)

  @parameterized.parameters(
    {'use_orbax': True, 'keep_every_n_steps': None},
    {'use_orbax': False, 'keep_every_n_steps': 7},
  )
  def test_keep(self, use_orbax, keep_every_n_steps):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    test_object = {'a': np.array([1, 2, 3], np.int32)}
    steps_start = 17
    steps_end = 37
    keep = 3
    increment = 5

    for step in range(steps_start, steps_end, increment):
      checkpoints.save_checkpoint(
        tmp_dir,
        test_object,
        step=step,
        keep=keep,
        keep_every_n_steps=keep_every_n_steps,
      )

    last_checkpoint = -float('inf')
    for step in range(steps_start, steps_end, increment):
      if ((steps_end - step) / increment <= keep) or (
        keep_every_n_steps and (step - last_checkpoint) >= keep_every_n_steps
      ):
        restored = checkpoints.restore_checkpoint(
          tmp_dir, target=None, step=step
        )
        check_eq(restored, test_object)
        last_checkpoint = step
      else:
        with self.assertRaises(ValueError):
          checkpoints.restore_checkpoint(tmp_dir, target=None, step=step)

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_save_restore_checkpoints_w_float_steps(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    test_object0 = {
      'a': np.array([0, 0, 0], np.int32),
      'b': np.array([0, 0, 0], np.int32),
    }
    test_object1 = {
      'a': np.array([1, 2, 3], np.int32),
      'b': np.array([1, 1, 1], np.int32),
    }
    test_object2 = {
      'a': np.array([4, 5, 6], np.int32),
      'b': np.array([2, 2, 2], np.int32),
    }
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 0.0, prefix='test_', keep=1
    )
    self.assertIn('test_0.0', os.listdir(tmp_dir))
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object0, prefix='test_'
    )
    check_eq(new_object, test_object1)
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 2.0, prefix='test_', keep=1
    )
    checkpoints.save_checkpoint(
      tmp_dir, test_object2, 3.0, prefix='test_', keep=2
    )
    self.assertIn('test_3.0', os.listdir(tmp_dir))
    self.assertIn('test_2.0', os.listdir(tmp_dir))
    check_eq(new_object, test_object1)

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_save_restore_checkpoints_target_none(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    test_object0 = {
      'a': np.array([0, 0, 0], np.int32),
      'b': np.array([0, 0, 0], np.int32),
    }
    # Target pytree is a dictionary, so it's equal to a restored state_dict.
    checkpoints.save_checkpoint(tmp_dir, test_object0, 0)
    new_object = checkpoints.restore_checkpoint(tmp_dir, target=None)
    check_eq(new_object, test_object0)
    # Target pytree it's a tuple, check the expected state_dict is recovered.
    test_object1 = (
      np.array([0, 0, 0], np.int32),
      np.array([1, 1, 1], np.int32),
    )
    checkpoints.save_checkpoint(tmp_dir, test_object1, 1)
    new_object = checkpoints.restore_checkpoint(tmp_dir, target=None)
    expected_new_object = {str(k): v for k, v in enumerate(test_object1)}
    check_eq(new_object, expected_new_object)

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_save_restore_checkpoints_target_singular(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    test_object0 = np.array([0, 0, 0], np.int32)
    test_object1 = np.array([1, 1, 1], np.int32)
    # Orbax backend returns error if target is singular. Orbax user need to use
    # ArrayCheckpointHandler instead.
    if use_orbax:
      with self.assertRaises(ValueError):
        checkpoints.save_checkpoint(tmp_dir, test_object1, 0)
    else:
      checkpoints.save_checkpoint(tmp_dir, test_object1, 0)
      new_object = checkpoints.restore_checkpoint(tmp_dir, target=None)
      check_eq(new_object, test_object1)
      checkpoints.save_checkpoint(tmp_dir, test_object0, 1)
      new_object = checkpoints.restore_checkpoint(tmp_dir, target=test_object1)
      check_eq(new_object, test_object0)

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_save_restore_checkpoints_target_empty(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    test_object0 = {}
    test_object1 = []
    # Orbax returns ValueError if the target is empty, but legacy Flax doesn't.
    if use_orbax:
      with self.assertRaises(ValueError):
        checkpoints.save_checkpoint(tmp_dir, test_object1, 0)
    else:
      checkpoints.save_checkpoint(tmp_dir, test_object1, 0)
      new_object = checkpoints.restore_checkpoint(tmp_dir, target=None)
      check_eq(new_object, test_object0)
      checkpoints.save_checkpoint(tmp_dir, test_object0, 1)
      new_object = checkpoints.restore_checkpoint(tmp_dir, target=test_object1)
      check_eq(new_object, test_object1)

  def test_async_save_checkpoints(self):
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    test_object0 = {
      'a': np.array([0, 0, 0], np.int32),
      'b': np.array([0, 0, 0], np.int32),
    }
    test_object1 = {
      'a': np.random.normal(size=(1000, 1000)),
      'b': np.random.normal(size=(1000, 1000)),
    }
    test_object2 = {
      'a': np.random.normal(size=(1000, 1000)),
      'b': np.random.normal(size=(1000, 1000)),
    }
    test_object3 = {
      'a': np.random.normal(size=(1000, 1000)),
      'b': np.random.normal(size=(1000, 1000)),
    }
    am = checkpoints.AsyncManager()
    checkpoints.save_checkpoint(
      tmp_dir, test_object1, 0, prefix='test_', keep=1, async_manager=am
    )
    # Hard-wait the write to be done, then check its content.
    am.save_future.result()
    self.assertIn('test_0', os.listdir(tmp_dir))
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object1, prefix='test_'
    )
    check_eq(new_object, test_object1)
    # Check two consecutive saves happen in the right order.
    checkpoints.save_checkpoint(
      tmp_dir, test_object2, 1, prefix='test_', keep=1, async_manager=am
    )
    checkpoints.save_checkpoint(
      tmp_dir, test_object3, 2, prefix='test_', keep=1, async_manager=am
    )
    am.save_future.result()
    self.assertIn('test_2', os.listdir(tmp_dir))
    new_object = checkpoints.restore_checkpoint(
      tmp_dir, test_object1, prefix='test_'
    )
    check_eq(new_object, test_object3)

  def test_last_checkpoint(self):
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    with io.GFile(os.path.join(tmp_dir, 'test_tmp'), 'w') as f:
      f.write('test_tmp')
    io.makedirs(os.path.join(tmp_dir, 'test_tmp_gda'))
    self.assertEqual(checkpoints.latest_checkpoint(tmp_dir, 'test_'), None)

    with io.GFile(os.path.join(tmp_dir, 'test_0'), 'w') as f:
      f.write('test_0')
    io.makedirs(os.path.join(tmp_dir, 'test_0_gda'))
    self.assertEqual(
      checkpoints.latest_checkpoint(tmp_dir, 'test_'),
      os.path.join(tmp_dir, 'test_0'),
    )

    with io.GFile(os.path.join(tmp_dir, 'test_10'), 'w') as f:
      f.write('test_10')
    self.assertEqual(
      checkpoints.latest_checkpoint(tmp_dir, 'test_'),
      os.path.join(tmp_dir, 'test_10'),
    )
    self.assertEqual(checkpoints.latest_checkpoint(tmp_dir, 'ckpt_'), None)

    path = f'orbaxtest_{orbax.utils.TMP_DIR_SUFFIX}_10'
    with io.GFile(os.path.join(tmp_dir, path), 'w') as f:
      f.write('orbaxtest_10')
    self.assertIsNone(checkpoints.latest_checkpoint(tmp_dir, 'orbaxtest_'))

  @parameterized.parameters(
    {'step_type': int, 'steps': [1, 5, 112]},
    {'step_type': float, 'steps': [1.0, 4.5, 5.6]},
  )
  def test_available_steps(self, step_type, steps):
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    with io.GFile(os.path.join(tmp_dir, 'test_tmp'), 'w') as f:
      f.write('test_tmp')
    io.makedirs(os.path.join(tmp_dir, 'test_tmp_gda'))

    for step in steps:
      with io.GFile(os.path.join(tmp_dir, 'test_' + str(step)), 'w') as f:
        f.write('test_' + str(step))
      io.makedirs(os.path.join(tmp_dir, 'test_' + str(step) + '_gda'))

    self.assertEqual(
      checkpoints.available_steps(tmp_dir, 'test_', step_type=step_type),
      steps,
    )

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_complex_pytree(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    to_save = [
      CustomDC(foo=12, bar=core.freeze({'x': jnp.array((1, 4))})),
      np.array((2, 3)),
    ]
    target = [
      CustomDC(foo=0, bar=core.freeze({'x': jnp.array((0, 0))})),
      np.array((0, 0)),
    ]
    checkpoints.save_checkpoint(tmp_dir, to_save, 0)
    restored = checkpoints.restore_checkpoint(tmp_dir, target=target)
    check_eq(restored, to_save)

  # restore_checkpoint can automatically restore either orbax or legacy files.
  def test_auto_restore(self):
    tmp_dir = self.create_tempdir().full_path
    to_save = [CustomDC(foo=12, bar={'x': jnp.array((1, 4))}), np.array((2, 3))]
    target = [CustomDC(foo=0, bar={'x': jnp.array((0, 0))}), np.array((0, 0))]
    # Store an orbax ckpt
    config.update('flax_use_orbax_checkpointing', True)
    checkpoints.save_checkpoint(tmp_dir, to_save, 0, prefix='test_')
    # And a legacy ckpt
    config.update('flax_use_orbax_checkpointing', False)
    checkpoints.save_checkpoint(tmp_dir, to_save, 1, prefix='test_', keep=2)

    # Both gets restored with same API.
    restored = checkpoints.restore_checkpoint(
      os.path.join(tmp_dir, 'test_0'), target=target
    )
    check_eq(restored, to_save)
    restored = checkpoints.restore_checkpoint(
      os.path.join(tmp_dir, 'test_1'), target=target
    )
    check_eq(restored, to_save)

  @parameterized.parameters({'use_orbax': True}, {'use_orbax': False})
  def test_smaller_target(self, use_orbax):
    config.update('flax_use_orbax_checkpointing', use_orbax)
    tmp_dir = self.create_tempdir().full_path
    to_save = {'a': jnp.ones((16, 256, 1024))}
    target = {'a': jnp.zeros((2, 3))}

    checkpoints.save_checkpoint(tmp_dir, to_save, 0, keep=1)
    new_object = checkpoints.restore_checkpoint(tmp_dir, target)
    check_eq(new_object, to_save)

  def test_convert_pre_linen(self):
    params = checkpoints.convert_pre_linen(
      {
        'mod_0': {
          'submod1_0': {},
          'submod2_1': {},
          'submod1_2': {},
        },
        'mod2_2': {'submod2_2_0': {}},
        'mod2_11': {'submod2_11_0': {}},
        'mod2_1': {'submod2_1_0': {}},
      }
    )
    self.assertDictEqual(
      core.unfreeze(params),
      {
        'mod_0': {
          'submod1_0': {},
          'submod1_1': {},
          'submod2_0': {},
        },
        'mod2_0': {'submod2_1_0': {}},
        'mod2_1': {'submod2_2_0': {}},
        'mod2_2': {'submod2_11_0': {}},
      },
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for sst2.train."""
from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.test_util
import numpy as np

from configs import default as default_config
import train


# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class TrainTest(parameterized.TestCase):

  def test_train_step_updates_parameters(self):
    """Tests if the train step updates the parameters in train state."""
    # Create model and a state that contains the parameters.
    config = default_config.get_config()
    config.vocab_size = 13
    rng = jax.random.key(config.seed)
    model = train.model_from_config(config)
    state = train.create_train_state(rng, config, model)

    token_ids = np.array([[2, 4, 3], [2, 6, 3]], dtype=np.int32)
    lengths = np.array([2, 3], dtype=np.int32)
    labels = np.zeros_like(lengths)
    batch = {'token_ids': token_ids, 'length': lengths, 'label': labels}
    rngs = {'dropout': rng}
    train_step_fn = jax.jit(train.train_step)
    new_state, metrics = train_step_fn(state, batch, rngs)
    self.assertIsInstance(new_state, train.TrainState)
    self.assertIsInstance(metrics, train.Metrics)
    old_param_values = jax.tree_util.tree_leaves(state.params)
    new_param_values = jax.tree_util.tree_leaves(new_state.params)
    for old_array, new_array in zip(old_param_values, new_param_values):
      # Make sure parameters were updated.
      self.assertFalse(np.allclose(old_array, new_array))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from copy import deepcopy
from typing import Any, TypeVar

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from flax import nnx

A = TypeVar('A')


class TestModule:
  def test_has_module_state(self):
    class Foo(nnx.Module): ...

    foo = Foo()

    assert hasattr(foo, '_object__state')

  def test_trace_level(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def f():
      with pytest.raises(
        nnx.errors.TraceContextError,
        match="Cannot mutate 'Dict' from different trace level",
      ):
        m.a = 2

    f()

  def test_tree_map(self):
    m = nnx.Dict(a=nnx.Param(1))

    graphdef, state = nnx.split(m)

    state = jax.tree_util.tree_map(lambda x: x + 1, state)

  def test_split_2(self):
    m = nnx.Dict(a=nnx.Param(1))

    graphdef, empty, some = nnx.split(m, None, ...)

    some = jax.tree_util.tree_map(lambda x: x + 1, some)

  def test_split_merge(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(graphdef: nnx.GraphDef[nnx.Dict[int]], state: nnx.State):
      m = nnx.merge(graphdef, state)
      m.a = 2
      return nnx.split(m)

    graphdef, state = g(*nnx.split(m))
    m2 = nnx.merge(graphdef, state)

    assert m2.a == 2

  def test_no_trace_level_error_on_grad(self):
    # No trace level error occurs because jax doesn't update
    # its top trace for grad.
    m = nnx.Dict(a=nnx.Param(1.0))

    @jax.grad
    def f(_):
      m.a = 2.0
      return 1.0

    f(1.0)

  def test_call(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, ()))
        self.c = c

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w.value * x + jax.random.normal(key, ()) + self.c

    foo = Foo(c=1.0, rngs=nnx.Rngs(0))

    y = foo(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_shared_module(self):
    m1 = nnx.Dict(a=nnx.Param(1), b=nnx.Param(2))
    m2 = nnx.Dict(x=m1, y=m1, z=nnx.Param(3))

    m3 = nnx.merge(*nnx.split(m2))

    assert m3['x'] is m3['y']
    assert m3['x']['a'] is m3['y']['a']
    assert m3['x']['b'] is m3['y']['b']

  def test_module_graph(self):
    class Foo(nnx.Module):
      def __init__(self):
        self.a = nnx.Param(1)
        self.sub = self

    m = Foo()

    graphdef, state = nnx.split(m)
    assert len(state) == 1

    m2 = nnx.merge(graphdef, state)
    assert m2 is m2.sub

  def test_deref_through_jit(self):
    r1 = nnx.Variable(1)
    r2 = nnx.Variable(2)

    m = m0 = nnx.Dict({'a': nnx.List([r1, r2]), 'b': r1})

    @jax.jit
    def f(graphdef: nnx.GraphDef[nnx.Dict[Any]], state: nnx.State):
      m = nnx.merge(graphdef, state)

      assert m['a'][0] is m['b']
      assert m['a'][1] is not m['b']

      return nnx.split(m)

    graphdef, state = f(*nnx.split(m))
    m = nnx.merge(graphdef, state)

    assert m['a'][0] is m['b']
    assert m['a'][1] is not m['b']

    # compare with original
    assert m['a'][0] is not m0['a'][0]
    assert m['a'][1] is not m0['a'][1]
    assert m['b'] is not m0['b']

  def test_cross_barrier(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(graphdef: nnx.GraphDef[nnx.Dict[nnx.Param[int]]], state: nnx.State):
      m = nnx.merge(graphdef, state)
      m.a.value += 1
      return nnx.split(m)

    graphdef, state = g(*nnx.split(m))
    m2 = nnx.merge(graphdef, state)
    assert m2 is not m
    assert m.a.value == 1
    assert m2.a.value == 2

  def test_no_rejit(self):
    n = 0
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(state_and_def):
      nonlocal n
      n += 1
      m = nnx.merge(*state_and_def)
      m.a.value += 1
      return nnx.split(m)

    m2 = nnx.merge(*g(nnx.split(m)))

    assert n == 1
    assert m2 is not m
    assert m.a.value == 1
    assert m2.a.value == 2

    g(nnx.split(m))
    assert n == 1

    g(nnx.split(m2))
    assert n == 1

    m2.b = nnx.Param(10)
    g(nnx.split(m2))

    assert n == 2

  def test_deref_number_of_fields(self):
    r1 = nnx.Variable(1)
    r2 = nnx.Variable(2)
    v1 = 3
    m = nnx.Dict(
      {
        'a': nnx.List([r1, r2, v1]),
        'b': nnx.Dict({'c': r1, 'd': r2}),
      }
    )

    graphdef, p = nnx.split(m)
    assert len(p.flat_state()) == 2
    assert len(jax.tree_util.tree_leaves(p)) == 2

  def test_clone(self):
    m = nnx.Dict(
      a=nnx.List([nnx.Param(1), nnx.Param(2), 3]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.Param(2)),
    )

    m2 = nnx.clone(m)

    assert m is not m2
    assert m2.a[0] == m2.b.c
    assert m2.a[1] == m2.b.d

    assert m.a[0] == m2.a[0]
    assert m.a[1] == m2.a[1]
    assert m.b.c == m2.b.c
    assert m.b.d == m2.b.d

  def test_sow_basic(self):
    class Foo(nnx.Module):
      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()
    y1 = m(2)
    y2 = m(10)

    assert y1 == 3
    assert y2 == 11
    assert m.y.value == (3, 11)

    intermediates = nnx.pop(m, nnx.Intermediate)

    assert issubclass(intermediates.y.type, nnx.Intermediate)
    assert intermediates['y'].value == (3, 11)

    assert not hasattr(m, 'y')

  def test_sow_existing_non_variable_field(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.y = 10

      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()

    with pytest.raises(ValueError, match='to be a Variable, got'):
      m(2)

  def test_sow_wrong_collection(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.y = nnx.Param(10)

      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()

    with pytest.raises(ValueError, match='to be of type'):
      m(2)

  def test_update_static_state_submodules(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

      def add_field(self):
        self.y = 2

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

    m1 = Foo()
    with nnx.update_context('test') as ctx:
      graphdef, state = ctx.split(m1)
      m2 = ctx.merge(graphdef, state)
      m2.a.add_field()
      new_graphdef, state = ctx.split(m2)

      m3 = ctx.merge(new_graphdef, state)

    assert m3 is m1
    assert m1.a.x == 1
    assert m1.a.y == 2
    assert m1.b.x == 1
    assert m1.b.y == 2

  def test_update_new_submodule(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()

      def add_module(self):
        self.b = Bar()

    m1 = Foo()
    with nnx.update_context('test') as ctx:
      graphdef, state = ctx.split(m1)
      m2 = ctx.merge(graphdef, state)
      m2.add_module()
      new_graphdef, state = ctx.split(m2)

      m3 = ctx.merge(new_graphdef, state)

    assert m3 is m1
    assert m1.a.x == 1
    assert m1.b.x == 1

  def test_update_update_submodule(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

    m1 = Foo()
    with nnx.update_context('test') as ctx:
      graphdef, state = ctx.split(m1)
      m2 = ctx.merge(graphdef, state)
      m2.a.x = 2
      new_graphdef, state = ctx.split(m2)
      m3 = ctx.merge(new_graphdef, state)

    assert m3 is m1
    assert m1.a.x == 2
    assert m1.b.x == 2

  def test_update_add_shared(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

      def add_submodule(self):
        self.c = self.a

    m1 = Foo()
    with nnx.update_context('test') as ctx:
      graphdef, state = ctx.split(m1)
      m2 = ctx.merge(graphdef, state)
      m2.add_submodule()
      new_graphdef, state = ctx.split(m2)
      m3 = ctx.merge(new_graphdef, state)

    assert m3 is m1
    assert hasattr(m1, 'c')

  def test_create_abstract(self):
    linear = nnx.eval_shape(lambda: nnx.Linear(2, 3, rngs=nnx.Rngs(0)))

    assert linear.kernel.value == jax.ShapeDtypeStruct((2, 3), jnp.float32)
    assert linear.bias.value == jax.ShapeDtypeStruct((3,), jnp.float32)

  def test_partial_init(self):
    linear = nnx.Linear(2, 3, rngs=nnx.Rngs(0))
    state = nnx.state(linear)

    del state['bias']

    @nnx.jit
    def partial_init(state: nnx.State):
      m = nnx.Linear(
        2, 3, bias_init=nnx.initializers.ones_init(), rngs=nnx.Rngs(1)
      )
      nnx.update(m, state)
      return m

    linear2 = partial_init(state)

    np.testing.assert_allclose(linear.kernel.value, linear2.kernel.value)
    np.testing.assert_allclose(linear.bias.value, 0)
    np.testing.assert_allclose(linear2.bias.value, 1)

  def test_deepcopy(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = nnx.Param(1)
        self.b = [1, 2, 3]
        self.c = nnx.Param(jnp.array([1.0]))
        self.self = self

    m1 = Foo()
    m2 = deepcopy(m1)

    assert m1.a == m2.a
    assert vars(m1)['a'] is not vars(m2)['a']
    assert m1.b is not m2.b
    assert m1.c is not m2.c
    assert m1.self is m1

  def test_set_attributes(self):
    class Block(nnx.Module):
      def __init__(self, din, dout, *, rngs: nnx.Rngs):
        self.linear = nnx.Linear(din, dout, rngs=rngs)
        self.dropout = nnx.Dropout(0.5, deterministic=False)
        self.batch_norm = nnx.BatchNorm(
          10, use_running_average=False, rngs=rngs
        )

    block = Block(2, 5, rngs=nnx.Rngs(0))
    assert block.dropout.deterministic == False
    assert block.batch_norm.use_running_average == False

    block.set_attributes(deterministic=True, use_running_average=True)
    assert block.dropout.deterministic == True
    assert block.batch_norm.use_running_average == True

    block = Block(2, 5, rngs=nnx.Rngs(0))
    block.set_attributes(nnx.Dropout, deterministic=True)
    # Only the dropout will be modified
    assert block.dropout.deterministic == True
    assert block.batch_norm.use_running_average == False

  def test_set_attribute_error(self):
    class Block(nnx.Module):
      def __init__(self, din, dout, *, rngs: nnx.Rngs):
        self.linear = nnx.Linear(din, dout, rngs=rngs)
        self.dropout = nnx.Dropout(0.5, deterministic=False)
        self.batch_norm = nnx.BatchNorm(
          10, use_running_average=False, rngs=rngs
        )

    block = Block(2, 5, rngs=nnx.Rngs(0))

    with pytest.raises(
      ValueError,
      match="Could not find at least one instance of the following attributes: {'unknown'}",
    ):
      block.set_attributes(
        deterministic=True, use_running_average=True, unknown=True
      )

    block.set_attributes(
      deterministic=True,
      use_running_average=True,
      unknown=True,
      raise_if_not_found=False,
    )


class TestModulePytree:
  def test_tree_map(self):
    class Foo(nnx.Module, experimental_pytree=True):
      def __init__(self):
        self.node = nnx.Param(1)
        self.graphdef = 1

    m = Foo()

    m = jax.tree_util.tree_map(lambda x: x + 1, m)

    assert m.node.value == 2
    assert m.graphdef == 1

  def test_static(self):
    class C(nnx.Module, experimental_pytree=True):
      def __init__(self, x):
        self.x = x

    n = 0

    @jax.jit
    def f(x):
      nonlocal n
      n += 1

    f(C(1))
    assert n == 1
    f(C(1))
    assert n == 1
    f(C(2))
    assert n == 2
    f(C(2))
    assert n == 2


class TestModuleDataclass:
  def test_basic(self):
    @dataclasses.dataclass
    class Foo(nnx.Module):
      a: int
      b: nnx.Variable[int]
      c: nnx.Param[int]
      d: nnx.Variable[int]
      e: nnx.Variable[int]
      f: int

    m = Foo(
      a=1,  # graphdef
      b=nnx.Variable(2),  # node
      c=nnx.Param(3),  # param
      d=nnx.Variable(4),  # var
      e=nnx.BatchStat(5),  # var
      f=6,  # graphdef int
    )

    graphdef, state = nnx.split(m)

    assert len(state) == 4
    assert state.b.value == 2
    assert state.b.type == nnx.Variable
    assert state.c.value == 3
    assert state.c.type == nnx.Param
    assert state.d.value == 4
    assert state.d.type == nnx.Variable
    assert state.e.value == 5
    assert state.e.type == nnx.BatchStat

  def test_post_init(self):
    @dataclasses.dataclass
    class DFoo(nnx.Module):
      din: int
      dout: int
      rngs: nnx.Rngs

      def __post_init__(self):
        self.bar = nnx.Linear(self.din, self.dout, rngs=self.rngs)

      def __call__(self, x):
        return self.bar(x)

    m = DFoo(1, 1, rngs=nnx.Rngs(0))

    assert hasattr(m, 'bar')


class TestModuleDef:
  def test_apply(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        self.w = nnx.Param(jax.random.uniform(rngs.params(), ()))
        self.c = c

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w.value * x + jax.random.normal(key, ()) + self.c

    rngs = nnx.Rngs(0)
    foo = Foo(c=1.0, rngs=rngs)

    graphdef, states = nnx.split(foo)

    assert isinstance(states, nnx.State)
    assert issubclass(states.w.type, nnx.Param)

    y, _updates = graphdef.apply(states)(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_derefed_mod_apply(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        self.w = nnx.Param(
          jax.random.uniform(rngs.params(), ()),
        )
        self.c = nnx.Variable(c)

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w.value * x + jax.random.normal(key, ()) + self.c.value

    foo = Foo(c=1.0, rngs=nnx.Rngs(0))

    graphdef, state = nnx.split(foo)

    assert isinstance(graphdef, nnx.GraphDef)
    assert isinstance(state, nnx.State)
    assert issubclass(state.w.type, nnx.Param)
    assert issubclass(state.c.type, nnx.Variable)

    y, (graphdef, state) = graphdef.apply(state)(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_modules_iterator(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        self.submodules = [
          {'a': nnx.Linear(1, 1, rngs=rngs)},
          {'b': nnx.Conv(1, 1, 1, rngs=rngs)},
        ]
        self.linear = nnx.Linear(1, 1, rngs=rngs)
        self.dropout = nnx.Dropout(0.5, rngs=rngs)

    module = Foo(rngs=nnx.Rngs(0))

    modules = list(module.iter_modules())

    assert len(modules) == 5
    assert modules[0][0] == ('dropout',)
    assert isinstance(modules[0][1], nnx.Dropout)
    assert modules[1][0] == ('linear',)
    assert isinstance(modules[1][1], nnx.Linear)
    assert modules[2][0] == ('submodules', 0, 'a')
    assert isinstance(modules[2][1], nnx.Linear)
    assert modules[3][0] == ('submodules', 1, 'b')
    assert isinstance(modules[3][1], nnx.Conv)
    assert modules[4][0] == ()
    assert isinstance(modules[4][1], Foo)

  def test_children_modules_iterator(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        self.submodules = [
          {'a': nnx.Linear(1, 1, rngs=rngs)},
          {'b': nnx.Conv(1, 1, 1, rngs=rngs)},
        ]
        self.linear = nnx.Linear(1, 1, rngs=rngs)
        self.dropout = nnx.Dropout(0.5, rngs=rngs)

    module = Foo(rngs=nnx.Rngs(0))

    modules = list(module.iter_children())

    assert len(modules) == 2
    assert modules[0][0] == 'dropout'
    assert isinstance(modules[0][1], nnx.Dropout)
    assert modules[1][0] == 'linear'
    assert isinstance(modules[1][1], nnx.Linear)

  def test_array_in_module(self):
    class Foo(nnx.Module):
      def __init__(self):
        self.a = jnp.array(1.0)

    foo = Foo()

    graphdef, state = nnx.split(foo)

    assert isinstance(state, nnx.State)
    assert isinstance(state.a, jax.Array)

    foo2 = nnx.merge(graphdef, state)

    assert isinstance(foo2.a, jax.Array)

  def test_state_in_module(self):
    class Foo(nnx.Module):
      def __init__(self):
        self.a = nnx.State({'b': jnp.array(1.0)})

    foo = Foo()

    graphdef, state = nnx.split(foo)

    assert isinstance(state, nnx.State)
    assert isinstance(state.a, nnx.State)

    foo2 = nnx.merge(graphdef, state)

    assert isinstance(foo2.a, nnx.State)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any


import dataclasses

import jax
import jax.numpy as jnp
from absl.testing import absltest
from jax import random

from flax import nnx
from configs import default
from models import TransformerConfig, TransformerLM
from utils import HasCache

jax.config.update('jax_disable_most_optimizations', True)

# add project_root to import lm1b Linen model
project_root = str(Path(__file__).absolute().parents[4])
sys.path.append(project_root)
from examples.lm1b.models import TransformerLM as TransformerLinen  # type: ignore[import-error]

sys.path.pop()


@dataclasses.dataclass(unsafe_hash=True)
class CompatTransformerConfig(TransformerConfig):
  decode: bool | None = None
  deterministic: bool | None = None


def get_transformer_config(**kwargs):
  base_config = default.get_config()
  config = CompatTransformerConfig(
    vocab_size=base_config.vocab_size,
    output_vocab_size=base_config.vocab_size,
    logits_via_embedding=base_config.logits_via_embedding,
    dtype=jnp.bfloat16 if base_config.use_bfloat16 else jnp.float32,
    emb_dim=base_config.emb_dim,
    num_heads=base_config.num_heads,
    num_layers=base_config.num_layers,
    qkv_dim=base_config.qkv_dim,
    mlp_dim=base_config.mlp_dim,
    max_len=max(
      base_config.max_target_length, base_config.max_eval_target_length
    ),
    dropout_rate=base_config.dropout_rate,
    attention_dropout_rate=base_config.attention_dropout_rate,
    kernel_init=nnx.initializers.xavier_uniform(),
    bias_init=nnx.initializers.normal(stddev=1e-6),
    **kwargs,
  )
  return base_config, config


class ModelTest(absltest.TestCase):
  def transfer_params(
    self,
    config: TransformerConfig,
    params_nnx: nnx.State,
    params_linen: dict[str, Any],
  ):
    rules = dataclasses.asdict(config.axis_rules)
    flat_params_nnx = params_nnx.flat_state()
    flat_params_linen = nnx.traversals.flatten_mapping(params_linen, sep='/')

    def apply_rules(names: tuple[str, ...]):
      return tuple(rules[name] for name in names)

    def copy_var(nnx_name: str, linen_name: str):
      nnx_path = tuple(nnx_name.split('/'))
      assert (
        flat_params_nnx[nnx_path].value.shape
        == flat_params_linen[linen_name].value.shape
      )
      flat_params_nnx[nnx_path].value = flat_params_linen[linen_name].value
      assert flat_params_nnx[nnx_path].sharding == apply_rules(
        flat_params_linen[linen_name].names
      )

    copy_var('decoder/output_embed/embedding', 'decoder/Embed_0/embedding')
    copy_var(
      'decoder/encoderdecoder_norm/bias', 'decoder/encoderdecoder_norm/bias'
    )
    copy_var(
      'decoder/encoderdecoder_norm/scale', 'decoder/encoderdecoder_norm/scale'
    )

    for idx in range(config.num_layers):
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/ln1/bias',
        f'decoder/encoderdecoderblock_{idx}/LayerNorm_0/bias',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/ln1/scale',
        f'decoder/encoderdecoderblock_{idx}/LayerNorm_0/scale',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/ln2/bias',
        f'decoder/encoderdecoderblock_{idx}/LayerNorm_1/bias',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/ln2/scale',
        f'decoder/encoderdecoderblock_{idx}/LayerNorm_1/scale',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/query/kernel',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/query/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/key/kernel',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/key/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/value/kernel',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/value/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/out/kernel',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/out/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/mlp/linear1/kernel',
        f'decoder/encoderdecoderblock_{idx}/MlpBlock_0/Dense_0/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/mlp/linear1/bias',
        f'decoder/encoderdecoderblock_{idx}/MlpBlock_0/Dense_0/bias',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/mlp/linear2/kernel',
        f'decoder/encoderdecoderblock_{idx}/MlpBlock_0/Dense_1/kernel',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/mlp/linear2/bias',
        f'decoder/encoderdecoderblock_{idx}/MlpBlock_0/Dense_1/bias',
      )

    copy_var('decoder/logitdense/kernel', 'decoder/logitdense/kernel')
    copy_var('decoder/logitdense/bias', 'decoder/logitdense/bias')

  def transfer_cache(
    self,
    config: TransformerConfig,
    cache_nnx: nnx.State,
    cache_linen: dict[str, Any],
  ):
    flat_cache_nnx = cache_nnx.flat_state()
    flat_cache_linen = nnx.traversals.flatten_mapping(cache_linen, sep='/')

    def copy_var(nnx_name: str, linen_name: str):
      nnx_path = tuple(nnx_name.split('/'))
      assert (
        flat_cache_nnx[nnx_path].value.shape
        == flat_cache_linen[linen_name].shape
      )
      flat_cache_nnx[nnx_path].value = flat_cache_linen[linen_name]

    for idx in range(config.num_layers):
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/cache_index',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/cache_index',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/cached_key',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/cached_key',
      )
      copy_var(
        f'decoder/encoderdecoderblock_{idx}/attention/cached_value',
        f'decoder/encoderdecoderblock_{idx}/MultiHeadDotProductAttention_0/cached_value',
      )

    copy_var(
      'decoder/posembed_output/cache_index',
      'decoder/posembed_output/cache_index',
    )

  def test_forward_eval(self):
    _, config = get_transformer_config(
      axis_rules=default.MeshRules(
        embed='model',
        mlp='data',
        kv=None,
        vocab=None,
      ),
      deterministic=True,
      decode=False,
    )
    # Set dropout rates to avoid create dropout states
    config.dropout_rate = 0.0
    config.attention_dropout_rate = 0.0

    model_nnx = nnx.eval_shape(lambda: TransformerLM(config, rngs=nnx.Rngs(0)))
    _, params_nnx = nnx.split(model_nnx, nnx.Param)

    model_linen = TransformerLinen(config)

    sample_inputs = random.randint(random.PRNGKey(0), (1, 3), 0, 20)
    params_linen = model_linen.init(random.key(0), sample_inputs)['params']

    self.transfer_params(config, params_nnx, params_linen)
    nnx.update(model_nnx, params_nnx)

    model_nnx.set_attributes(deterministic=True, decode=False)
    output_nnx = model_nnx(sample_inputs)

    output_linen: jax.Array = model_linen.apply(
      {'params': params_linen}, sample_inputs
    )

    assert jnp.allclose(output_nnx, output_linen, atol=1e-5)

  def test_forward_decode(self):
    batch_size = 2

    _, config = get_transformer_config(
      axis_rules=default.MeshRules(
        embed='model',
        mlp='data',
        kv=None,
        vocab=None,
      ),
      deterministic=True,
      decode=True,
    )
    # Set dropout rates to avoid create dropout states
    config.dropout_rate = 0.0
    config.attention_dropout_rate = 0.0

    model_nnx = nnx.eval_shape(lambda: TransformerLM(config, rngs=nnx.Rngs(0)))
    for _path, m in model_nnx.iter_modules():
      if isinstance(m, HasCache):
        input_shape = (batch_size, config.max_len, config.emb_dim)
        m.init_cache(input_shape, dtype=config.dtype)

    _, params_nnx, cache_nnx = nnx.split(model_nnx, nnx.Param, nnx.Cache)

    model_linen = TransformerLinen(config)

    flax_init_inputs = random.randint(
      random.PRNGKey(0), (batch_size, config.max_len), 0, config.vocab_size
    )
    ar_decode_inputs = random.randint(
      random.PRNGKey(0), (3, batch_size, 1), 0, config.vocab_size
    )
    variables = model_linen.init(random.key(0), flax_init_inputs)
    params_linen = variables['params']
    cache_linen = variables['cache']

    self.transfer_params(config, params_nnx, params_linen)
    self.transfer_cache(config, cache_nnx, cache_linen)
    nnx.update(model_nnx, params_nnx, cache_nnx)
    model_nnx.set_attributes(deterministic=True, decode=True)

    outputs_nnx = []
    outputs_linen = []

    for inputs in ar_decode_inputs:
      output_nnx = model_nnx(inputs)
      outputs_nnx.append(output_nnx)

    output_linen: jax.Array
    for inputs in ar_decode_inputs:
      output_linen, updates = model_linen.apply(
        {'params': params_linen, 'cache': cache_linen},
        inputs,
        mutable=['cache'],
      )
      cache_linen = updates['cache']
      outputs_linen.append(output_linen)

    for output_nnx, output_linen in zip(outputs_nnx, outputs_linen):
      assert jnp.allclose(output_nnx, output_linen, atol=1e-5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dynamic loss scaling for mixed precision gradients."""

import functools
from typing import Any, Callable, NamedTuple, Optional, Sequence, Union

import jax
import jax.numpy as jnp
from jax import lax

from flax import struct
from flax.typing import Array


class DynamicScaleResult(NamedTuple):
  dynamic_scale: 'DynamicScale'
  finite: Array
  aux: Any
  grad: Any


class DynamicScale(struct.PyTreeNode):
  """Dynamic loss scaling for mixed precision gradients.

  For many models gradient computations in float16 will result in numerical
  issues because small/large gradients being flushed to zero/infinity.
  Dynamic loss scaling is an algorithm that aims to find the largest scalar
  multiple for which the gradient does not overflow. This way the risk of
  underflow is minimized.

  the `value_and_grad` method mimicks `jax.value_and_grad`. Beside the loss
  and gradients it also ouputs and updated `DynamicScale` instance with the
  current loss scale factor. This method also returns a boolean value indicating
  whether the gradients are finite.

  Example::

    from flax.training.dynamic_scale import DynamicScale

    def loss_fn(p):
      return jnp.asarray(p, jnp.float16) ** 2
    p = jnp.array(1., jnp.float32)

    dyn_scale = DynamicScale(growth_interval=10)
    compute_grad = jax.jit(lambda ds, p: ds.value_and_grad(loss_fn)(p))
    for _ in range(100):
      dyn_scale, is_fin, loss, grad = compute_grad(dyn_scale, p)
      p += jnp.where(is_fin, 0.01 * grad, 0.)
      print(loss)

  Jax currently cannot execute conditionals efficiently on GPUs therefore we
  selectively ignore the gradient update using `jax.numpy.where` in case of
  non-finite gradients.

  Attributes:
    growth_factor: how much to grow the scalar after a period of finite
      gradients (default: 2.).
    backoff_factor: how much to shrink the scalar after a non-finite gradient
      (default: 0.5).
    growth_interval: after how many steps of finite gradients the scale should
      be increased (default: 2000).
    fin_steps: indicates how many gradient steps in a row have been finite.
    scale: the current scale by which the loss is multiplied.
    minimum_scale: the minimum value that the scale can take (default: the
      smallest positive number representable in floating point).
  """

  growth_factor: float = struct.field(pytree_node=False, default=2.0)
  backoff_factor: float = struct.field(pytree_node=False, default=0.5)
  growth_interval: int = struct.field(pytree_node=False, default=2000)
  fin_steps: int = 0
  scale: float = 65536.0
  minimum_scale: Optional[float] = struct.field(
    pytree_node=False, default=jnp.finfo(jnp.float32).tiny
  )

  def value_and_grad(
    self,
    fun: Callable[..., Any],
    argnums: Union[int, Sequence[int]] = 0,
    has_aux: bool = False,
    axis_name: Optional[str] = None,
  ) -> Callable[..., DynamicScaleResult]:
    """Wrapper around `jax.value_and_grad`.

    Args:
      fun: Function to be differentiated. Its arguments at positions specified
        by ``argnums`` should be arrays, scalars, or standard Python containers.
        It should return a scalar (which includes arrays with shape ``()`` but
        not arrays with shape ``(1,)`` etc.)
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default 0).
      has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where
        the first element is considered the output of the mathematical function
        to be differentiated and the second element is auxiliary data. Default
        False.
      axis_name: If an axis is given the gradients will be averaged across
        replicas (default: None). Note, this is only used for pmap and shard
        map. For SPMD jit, you do not need to manually synchronize. Just make
        sure that the axes are correctly annotated and XLA:SPMD will insert the
        necessary collectives.

    Returns:
      A function that takes the same arguments as `fun` and
      returns a DynamicScaleResult
    """

    @functools.wraps(fun)
    def loss_wrapper(*args):
      aux = fun(*args)
      if has_aux:
        return (self.scale * aux[0], aux[1])
      else:
        return self.scale * aux

    grad_fn = jax.value_and_grad(loss_wrapper, argnums, has_aux)

    def grad_fn_wrapper(*args):
      aux, grad = grad_fn(*args)
      aux = (aux[0] / self.scale, aux[1]) if has_aux else aux / self.scale

      grad = jax.tree_util.tree_map(
        lambda g: jnp.asarray(g, jnp.float32) / self.scale, grad
      )
      if axis_name is not None:
        grad = lax.pmean(grad, axis_name)

      finite = jnp.array(True)
      for g in jax.tree_util.tree_leaves(grad):
        finite &= jnp.all(lax.is_finite(g))

      grow = self.fin_steps == self.growth_interval
      fin_scale = jnp.where(
        grow & finite,
        jnp.minimum(
          self.scale * self.growth_factor, jnp.finfo(jnp.float32).max
        ),
        self.scale,
      )
      inf_scale = self.scale * self.backoff_factor
      if self.minimum_scale is not None:
        inf_scale = jnp.maximum(inf_scale, self.minimum_scale)
      new_scale = jnp.where(finite, fin_scale, inf_scale)
      new_fin_steps = jnp.where(grow | (~finite), 0, self.fin_steps + 1)

      new_self = self.replace(fin_steps=new_fin_steps, scale=new_scale)
      return DynamicScaleResult(new_self, finite, aux, grad)

    return grad_fn_wrapper



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stochastic modules."""

from typing import Optional, Sequence

import jax.numpy as jnp
from jax import lax, random

from flax.linen.module import Module, compact, merge_param
from flax.typing import PRNGKey


class Dropout(Module):
  """Create a dropout layer.

  .. note::
    When using :meth:`Module.apply() <flax.linen.Module.apply>`, make sure
    to include an RNG seed named ``'dropout'``. Dropout isn't necessary for
    variable initialization.

  Example usage::

    >>> import flax.linen as nn
    >>> import jax, jax.numpy as jnp

    >>> class MLP(nn.Module):
    ...   @nn.compact
    ...   def __call__(self, x, train):
    ...     x = nn.Dense(4)(x)
    ...     x = nn.Dropout(0.5, deterministic=not train)(x)
    ...     return x

    >>> model = MLP()
    >>> x = jnp.ones((1, 3))
    >>> variables = model.init(jax.random.key(0), x, train=False) # don't use dropout
    >>> model.apply(variables, x, train=False) # don't use dropout
    Array([[-0.88686204, -0.5928178 , -0.5184689 , -0.4345976 ]], dtype=float32)
    >>> model.apply(variables, x, train=True, rngs={'dropout': jax.random.key(1)}) # use dropout
    Array([[ 0.       , -1.1856356, -1.0369378,  0.       ]], dtype=float32)

  Attributes:
    rate: the dropout probability.  (_not_ the keep rate!)
    broadcast_dims: dimensions that will share the same dropout mask
    deterministic: if false the inputs are scaled by ``1 / (1 - rate)`` and
      masked, whereas if true, no mask is applied and the inputs are returned as
      is.
    rng_collection: the rng collection name to use when requesting an rng key.
  """

  rate: float
  broadcast_dims: Sequence[int] = ()
  deterministic: Optional[bool] = None
  rng_collection: str = 'dropout'

  @compact
  def __call__(
    self,
    inputs,
    deterministic: Optional[bool] = None,
    rng: Optional[PRNGKey] = None,
  ):
    """Applies a random dropout mask to the input.

    Args:
      inputs: the inputs that should be randomly masked.
      deterministic: if false the inputs are scaled by ``1 / (1 - rate)`` and
        masked, whereas if true, no mask is applied and the inputs are returned
        as is.
      rng: an optional PRNGKey used as the random key, if not specified, one
        will be generated using ``make_rng`` with the ``rng_collection`` name.

    Returns:
      The masked inputs reweighted to preserve mean.
    """
    deterministic = merge_param(
      'deterministic', self.deterministic, deterministic
    )

    if (self.rate == 0.0) or deterministic:
      return inputs

    # Prevent gradient NaNs in 1.0 edge-case.
    if self.rate == 1.0:
      return jnp.zeros_like(inputs)

    keep_prob = 1.0 - self.rate
    if rng is None:
      rng = self.make_rng(self.rng_collection)
    broadcast_shape = list(inputs.shape)
    for dim in self.broadcast_dims:
      broadcast_shape[dim] = 1
    mask = random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
    mask = jnp.broadcast_to(mask, inputs.shape)
    return lax.select(mask, inputs / keep_prob, jnp.zeros_like(inputs))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

import jax
import jax.numpy as jnp

from absl.testing import absltest
from flax import nnx

A = tp.TypeVar('A')


class TestVariableState(absltest.TestCase):
  def test_pytree(self):
    r1 = nnx.VariableState(nnx.Param, 1)
    assert r1.value == 1

    r2 = jax.tree_util.tree_map(lambda x: x + 1, r1)

    assert r1.value == 1
    assert r2.value == 2
    assert r1 is not r2

  def test_overloads_module(self):
    class Linear(nnx.Module):
      def __init__(self, din, dout, rngs: nnx.Rngs):
        key = rngs()
        self.w = nnx.Param(jax.random.normal(key, (din, dout)))
        self.b = nnx.Param(jax.numpy.zeros((dout,)))

      def __call__(self, x: jax.Array):
        return x @ self.w + self.b

    linear = Linear(3, 4, nnx.Rngs(0))
    x = jax.numpy.ones((3,))
    y = linear(x)
    assert y.shape == (4,)

  def test_jax_array(self):
    class Linear(nnx.Module):
      def __init__(self, din, dout, rngs: nnx.Rngs):
        key = rngs()
        self.w = nnx.Param(jax.random.normal(key, (din, dout)))
        self.b = nnx.Param(jax.numpy.zeros((dout,)))

      def __call__(self, x: jax.Array):
        return jnp.dot(x, self.w) + self.b  # type: ignore[arg-type]

    linear = Linear(3, 4, nnx.Rngs(0))
    x = jax.numpy.ones((3,))
    y = linear(x)
    assert y.shape == (4,)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformer-based machine translation model."""

# pylint: disable=attribute-defined-outside-init,g-bare-generic
# See issue #620.
# pytype: disable=wrong-arg-count
# pytype: disable=wrong-keyword-args
# pytype: disable=attribute-error

from typing import Callable, Any, Optional

from flax import linen as nn
from flax import struct
from jax import lax
import jax.numpy as jnp
import numpy as np


@struct.dataclass
class TransformerConfig:
  """Global hyperparameters used to minimize obnoxious kwarg plumbing."""

  vocab_size: int
  output_vocab_size: int
  share_embeddings: bool = False
  logits_via_embedding: bool = False
  dtype: Any = jnp.float32
  emb_dim: int = 512
  num_heads: int = 8
  num_layers: int = 6
  qkv_dim: int = 512
  mlp_dim: int = 2048
  max_len: int = 2048
  dropout_rate: float = 0.1
  attention_dropout_rate: float = 0.1
  deterministic: bool = False
  decode: bool = False
  kernel_init: Callable = nn.initializers.xavier_uniform()
  bias_init: Callable = nn.initializers.normal(stddev=1e-6)
  posemb_init: Optional[Callable] = None


def shift_right(x, axis=1):
  """Shift the input to the right by padding on axis 1."""
  pad_widths = [(0, 0)] * len(x.shape)
  pad_widths[axis] = (1, 0)
  padded = jnp.pad(
      x, pad_widths, mode='constant', constant_values=x.dtype.type(0)
  )
  return padded[:, :-1]


def sinusoidal_init(max_len=2048, min_scale=1.0, max_scale=10000.0):
  """1D Sinusoidal Position Embedding Initializer.

  Args:
      max_len: maximum possible length for the input.
      min_scale: float: minimum frequency-scale in sine grating.
      max_scale: float: maximum frequency-scale in sine grating.

  Returns:
      output: init function returning `(1, max_len, d_feature)`
  """

  def init(key, shape, dtype=np.float32):
    """Sinusoidal init."""
    del key, dtype
    d_feature = shape[-1]
    pe = np.zeros((max_len, d_feature), dtype=np.float32)
    position = np.arange(0, max_len)[:, np.newaxis]
    scale_factor = -np.log(max_scale / min_scale) / (d_feature // 2 - 1)
    div_term = min_scale * np.exp(np.arange(0, d_feature // 2) * scale_factor)
    pe[:, : d_feature // 2] = np.sin(position * div_term)
    pe[:, d_feature // 2 : 2 * (d_feature // 2)] = np.cos(position * div_term)
    pe = pe[np.newaxis, :, :]  # [1, max_len, d_feature]
    return jnp.array(pe)

  return init


class AddPositionEmbs(nn.Module):
  """Adds (optionally learned) positional embeddings to the inputs.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
    decode: whether to run in single-position autoregressive mode.
  """

  config: TransformerConfig
  decode: bool = False

  @nn.compact
  def __call__(self, inputs, inputs_positions=None):
    """Applies AddPositionEmbs module.

    By default this layer uses a fixed sinusoidal embedding table. If a
    learned position embedding is desired, pass an initializer to
    posemb_init in the configuration.

    Args:
      inputs: input data.
      inputs_positions: input position indices for packed sequences.

    Returns:
      output: `(bs, timesteps, in_dim)`
    """
    config = self.config
    # inputs.shape is (batch_size, seq_len, emb_dim)
    assert inputs.ndim == 3, (
        'Number of dimensions should be 3, but it is: %d' % inputs.ndim
    )
    length = inputs.shape[1]
    pos_emb_shape = (1, config.max_len, inputs.shape[-1])
    if config.posemb_init is None:
      # Use a fixed (non-learned) sinusoidal position embedding.
      pos_embedding = sinusoidal_init(max_len=config.max_len)(
          None, pos_emb_shape, None
      )
    else:
      pos_embedding = self.param(
          'pos_embedding', config.posemb_init, pos_emb_shape
      )
    pe = pos_embedding[:, :length, :]

    # We use a cache position index for tracking decoding position.
    if self.decode:
      is_initialized = self.has_variable('cache', 'cache_index')
      cache_index = self.variable(
          'cache', 'cache_index', lambda: jnp.array(0, dtype=jnp.uint32)
      )
      if is_initialized:
        i = cache_index.value
        cache_index.value = i + 1
        _, _, df = pos_embedding.shape
        pe = lax.dynamic_slice(pos_embedding, jnp.array((0, i, 0)), (1, 1, df))
    if inputs_positions is None:
      # normal unpacked case:
      return inputs + pe
    else:
      # for packed data we need to use known position indices:
      return inputs + jnp.take(pe[0], inputs_positions, axis=0)


class MlpBlock(nn.Module):
  """Transformer MLP / feed-forward block.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
    out_dim: optionally specify out dimension.
  """

  config: TransformerConfig
  out_dim: Optional[int] = None

  @nn.compact
  def __call__(self, inputs):
    """Applies Transformer MlpBlock module."""
    config = self.config
    actual_out_dim = inputs.shape[-1] if self.out_dim is None else self.out_dim
    x = nn.Dense(
        config.mlp_dim,
        dtype=config.dtype,
        kernel_init=config.kernel_init,
        bias_init=config.bias_init,
    )(inputs)
    x = nn.relu(x)
    x = nn.Dropout(rate=config.dropout_rate)(
        x, deterministic=config.deterministic
    )
    output = nn.Dense(
        actual_out_dim,
        dtype=config.dtype,
        kernel_init=config.kernel_init,
        bias_init=config.bias_init,
    )(x)
    output = nn.Dropout(rate=config.dropout_rate)(
        output, deterministic=config.deterministic
    )
    return output


class Encoder1DBlock(nn.Module):
  """Transformer encoder layer.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
  """

  config: TransformerConfig

  @nn.compact
  def __call__(self, inputs, encoder_mask=None):
    """Applies Encoder1DBlock module.

    Args:
      inputs: input data.
      encoder_mask: encoder self-attention mask.

    Returns:
      output after transformer encoder block.
    """
    config = self.config

    # Attention block.
    assert inputs.ndim == 3
    x = nn.LayerNorm(dtype=config.dtype)(inputs)
    x = nn.MultiHeadDotProductAttention(
        num_heads=config.num_heads,
        dtype=config.dtype,
        qkv_features=config.qkv_dim,
        kernel_init=config.kernel_init,
        bias_init=config.bias_init,
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=config.attention_dropout_rate,
        deterministic=config.deterministic,
    )(x, mask=encoder_mask)

    x = nn.Dropout(rate=config.dropout_rate)(
        x, deterministic=config.deterministic
    )
    x = x + inputs

    # MLP block.
    y = nn.LayerNorm(dtype=config.dtype)(x)
    y = MlpBlock(config=config)(y)

    return x + y


class EncoderDecoder1DBlock(nn.Module):
  """Transformer encoder-decoder layer.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
  """

  config: TransformerConfig

  @nn.compact
  def __call__(
      self, targets, encoded, decoder_mask=None, encoder_decoder_mask=None
  ):
    """Applies EncoderDecoder1DBlock module.

    Args:
      targets: input data for decoder
      encoded: input data from encoder
      decoder_mask: decoder self-attention mask.
      encoder_decoder_mask: encoder-decoder attention mask.

    Returns:
      output after transformer encoder-decoder block.
    """
    config = self.config

    # Decoder block.
    assert targets.ndim == 3
    x = nn.LayerNorm(dtype=config.dtype)(targets)
    x = nn.MultiHeadDotProductAttention(
        num_heads=config.num_heads,
        dtype=config.dtype,
        qkv_features=config.qkv_dim,
        kernel_init=config.kernel_init,
        bias_init=config.bias_init,
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=config.attention_dropout_rate,
        deterministic=config.deterministic,
        decode=config.decode,
    )(x, mask=decoder_mask)
    x = nn.Dropout(rate=config.dropout_rate)(
        x, deterministic=config.deterministic
    )
    x = x + targets

    # Encoder-Decoder block.
    y = nn.LayerNorm(dtype=config.dtype)(x)
    y = nn.MultiHeadDotProductAttention(
        num_heads=config.num_heads,
        dtype=config.dtype,
        qkv_features=config.qkv_dim,
        kernel_init=config.kernel_init,
        bias_init=config.bias_init,
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=config.attention_dropout_rate,
        deterministic=config.deterministic,
    )(y, encoded, mask=encoder_decoder_mask)

    y = nn.Dropout(rate=config.dropout_rate)(
        y, deterministic=config.deterministic
    )
    y = y + x

    # MLP block.
    z = nn.LayerNorm(dtype=config.dtype)(y)
    z = MlpBlock(config=config)(z)

    return y + z


class Encoder(nn.Module):
  """Transformer Model Encoder for sequence to sequence translation.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
    shared_embedding: a shared embedding layer to use.
  """

  config: TransformerConfig
  shared_embedding: Any = None

  @nn.compact
  def __call__(self, inputs, inputs_positions=None, encoder_mask=None):
    """Applies Transformer model on the inputs.

    Args:
      inputs: input data
      inputs_positions: input subsequence positions for packed examples.
      encoder_mask: decoder self-attention mask.

    Returns:
      output of a transformer encoder.
    """
    config = self.config
    assert inputs.ndim == 2  # (batch, len)

    # Input Embedding
    if self.shared_embedding is None:
      input_embed = nn.Embed(
          num_embeddings=config.vocab_size,
          features=config.emb_dim,
          embedding_init=nn.initializers.normal(stddev=1.0),
      )
    else:
      input_embed = self.shared_embedding
    x = inputs.astype('int32')
    x = input_embed(x)
    x = AddPositionEmbs(config=config, decode=False, name='posembed_input')(
        x, inputs_positions=inputs_positions
    )
    x = nn.Dropout(rate=config.dropout_rate)(
        x, deterministic=config.deterministic
    )

    x = x.astype(config.dtype)

    # Input Encoder
    for lyr in range(config.num_layers):
      x = Encoder1DBlock(config=config, name=f'encoderblock_{lyr}')(
          x, encoder_mask
      )

    encoded = nn.LayerNorm(dtype=config.dtype, name='encoder_norm')(x)

    return encoded


class Decoder(nn.Module):
  """Transformer Model Decoder for sequence to sequence translation.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
    shared_embedding: a shared embedding layer to use.
  """

  config: TransformerConfig
  shared_embedding: Any = None

  @nn.compact
  def __call__(
      self,
      encoded,
      targets,
      targets_positions=None,
      decoder_mask=None,
      encoder_decoder_mask=None,
  ):
    """Applies Transformer model on the inputs.

    Args:
      encoded: encoded input data from encoder.
      targets: target inputs.
      targets_positions: input subsequence positions for packed examples.
      decoder_mask: decoder self-attention mask.
      encoder_decoder_mask: encoder-decoder attention mask.

    Returns:
      output of a transformer decoder.
    """
    config = self.config

    assert encoded.ndim == 3  # (batch, len, depth)
    assert targets.ndim == 2  # (batch, len)

    # Target Embedding
    if self.shared_embedding is None:
      output_embed = nn.Embed(
          num_embeddings=config.output_vocab_size,
          features=config.emb_dim,
          embedding_init=nn.initializers.normal(stddev=1.0),
      )
    else:
      output_embed = self.shared_embedding

    y = targets.astype('int32')
    if not config.decode:
      y = shift_right(y)
    y = output_embed(y)
    y = AddPositionEmbs(
        config=config, decode=config.decode, name='posembed_output'
    )(y, inputs_positions=targets_positions)
    y = nn.Dropout(rate=config.dropout_rate)(
        y, deterministic=config.deterministic
    )

    y = y.astype(config.dtype)

    # Target-Input Decoder
    for lyr in range(config.num_layers):
      y = EncoderDecoder1DBlock(
          config=config, name=f'encoderdecoderblock_{lyr}'
      )(
          y,
          encoded,
          decoder_mask=decoder_mask,
          encoder_decoder_mask=encoder_decoder_mask,
      )
    y = nn.LayerNorm(dtype=config.dtype, name='encoderdecoder_norm')(y)

    # Decoded Logits
    if config.logits_via_embedding:
      # Use the transpose of embedding matrix for logit transform.
      logits = output_embed.attend(y.astype(jnp.float32))
      # Correctly normalize pre-softmax logits for this shared case.
      logits = logits / jnp.sqrt(y.shape[-1])
    else:
      logits = nn.Dense(
          config.output_vocab_size,
          dtype=config.dtype,
          kernel_init=config.kernel_init,
          bias_init=config.bias_init,
          name='logitdense',
      )(y)
    return logits


class Transformer(nn.Module):
  """Transformer Model for sequence to sequence translation.

  Attributes:
    config: TransformerConfig dataclass containing hyperparameters.
  """

  config: TransformerConfig

  def setup(self):
    config = self.config

    if config.share_embeddings:
      if config.output_vocab_size is not None:
        assert (
            config.output_vocab_size == config.vocab_size
        ), "can't share embedding with different vocab sizes."
      self.shared_embedding = nn.Embed(
          num_embeddings=config.vocab_size,
          features=config.emb_dim,
          embedding_init=nn.initializers.normal(stddev=1.0),
      )
    else:
      self.shared_embedding = None

    self.encoder = Encoder(
        config=config, shared_embedding=self.shared_embedding
    )
    self.decoder = Decoder(
        config=config, shared_embedding=self.shared_embedding
    )

  def encode(self, inputs, inputs_positions=None, inputs_segmentation=None):
    """Applies Transformer encoder-branch on the inputs.

    Args:
      inputs: input data.
      inputs_positions: input subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.

    Returns:
      encoded feature array from the transformer encoder.
    """
    config = self.config
    # Make padding attention mask.
    encoder_mask = nn.make_attention_mask(
        inputs > 0, inputs > 0, dtype=config.dtype
    )
    # Add segmentation block-diagonal attention mask if using segmented data.
    if inputs_segmentation is not None:
      encoder_mask = nn.combine_masks(
          encoder_mask,
          nn.make_attention_mask(
              inputs_segmentation,
              inputs_segmentation,
              jnp.equal,
              dtype=config.dtype,
          ),
      )
    return self.encoder(
        inputs, inputs_positions=inputs_positions, encoder_mask=encoder_mask
    )

  def decode(
      self,
      encoded,
      inputs,  # only needed for masks
      targets,
      targets_positions=None,
      inputs_segmentation=None,
      targets_segmentation=None,
  ):
    """Applies Transformer decoder-branch on encoded-input and target.

    Args:
      encoded: encoded input data from encoder.
      inputs: input data (only needed for masking).
      targets: target data.
      targets_positions: target subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.
      targets_segmentation: target segmentation info for packed examples.

    Returns:
      logits array from transformer decoder.
    """
    config = self.config

    # Make padding attention masks.
    if config.decode:
      # for fast autoregressive decoding only a special encoder-decoder mask is used
      decoder_mask = None
      encoder_decoder_mask = nn.make_attention_mask(
          jnp.ones_like(targets) > 0, inputs > 0, dtype=config.dtype
      )
    else:
      decoder_mask = nn.combine_masks(
          nn.make_attention_mask(targets > 0, targets > 0, dtype=config.dtype),
          nn.make_causal_mask(targets, dtype=config.dtype),
      )
      encoder_decoder_mask = nn.make_attention_mask(
          targets > 0, inputs > 0, dtype=config.dtype
      )

    # Add segmentation block-diagonal attention masks if using segmented data.
    if inputs_segmentation is not None:
      decoder_mask = nn.combine_masks(
          decoder_mask,
          nn.make_attention_mask(
              targets_segmentation,
              targets_segmentation,
              jnp.equal,
              dtype=config.dtype,
          ),
      )
      encoder_decoder_mask = nn.combine_masks(
          encoder_decoder_mask,
          nn.make_attention_mask(
              targets_segmentation,
              inputs_segmentation,
              jnp.equal,
              dtype=config.dtype,
          ),
      )
    logits = self.decoder(
        encoded,
        targets,
        targets_positions=targets_positions,
        decoder_mask=decoder_mask,
        encoder_decoder_mask=encoder_decoder_mask,
    )
    return logits.astype(self.config.dtype)

  def __call__(
      self,
      inputs,
      targets,
      inputs_positions=None,
      targets_positions=None,
      inputs_segmentation=None,
      targets_segmentation=None,
  ):
    """Applies Transformer model on the inputs.

    Args:
      inputs: input data.
      targets: target data.
      inputs_positions: input subsequence positions for packed examples.
      targets_positions: target subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.
      targets_segmentation: target segmentation info for packed examples.

    Returns:
      logits array from full transformer.
    """
    encoded = self.encode(
        inputs,
        inputs_positions=inputs_positions,
        inputs_segmentation=inputs_segmentation,
    )

    return self.decode(
        encoded,
        inputs,  # only used for masks
        targets,
        targets_positions=targets_positions,
        inputs_segmentation=inputs_segmentation,
        targets_segmentation=targets_segmentation,
    )



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Taken from flax/core/tracer.py 🏴‍☠️

import typing as tp

import jax
import jax.core
from jax.core import MainTrace

from flax.nnx.nnx import reprlib


@tp.runtime_checkable
class Tracer(tp.Protocol):
  _trace: jax.core.Trace


def get_top_trace(pytree: tp.Union[tp.Any, Tracer]) -> MainTrace:
  """Returns the main top trace of a sequence of tracers."""
  if isinstance(pytree, Tracer):
    return pytree._trace.main

  return jax.core.find_top_trace(jax.tree_util.tree_leaves(pytree)).main


def current_jax_trace() -> MainTrace:
  """Returns the innermost Jax tracer."""
  return get_top_trace(())


class TraceState(reprlib.Representable):
  __slots__ = ['_jax_trace']

  def __init__(self):
    self._jax_trace = current_jax_trace()

  @property
  def jax_trace(self):
    return self._jax_trace

  def is_valid(self) -> bool:
    return self._jax_trace is current_jax_trace()

  def __nnx_repr__(self):
    yield reprlib.Object(f'{type(self).__name__}')
    yield reprlib.Attr('jax_trace', self._jax_trace)

  def __penzai_repr__(self, path, subtree_renderer):
    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
    return pz_repr_lib.render_object_constructor(
        object_type=type(self),
        attributes={'jax_trace': self._jax_trace},
        path=path,
        subtree_renderer=subtree_renderer,
    )

  def __eq__(self, other):
    return isinstance(other, TraceState) and self._jax_trace is other._jax_trace



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.struct."""


import dataclasses
from typing import Any, NamedTuple

import jax
import jax.numpy as jnp
import optax
from absl.testing import absltest

import flax
import flax.linen as nn
from flax.core import freeze
from flax.cursor import AccessType, _traverse_tree, cursor
from flax.errors import CursorFindError, TraverseTreeError
from flax.training import train_state

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class GenericTuple(NamedTuple):
  x: Any
  y: Any = None
  z: Any = None


@dataclasses.dataclass
class GenericDataClass:
  x: Any
  y: Any = None
  z: Any = None


class CursorTest(absltest.TestCase):
  def test_repr(self):
    g = GenericTuple(1, 'a', (2, 'b'))
    c = cursor(
      {'a': {1: {(2, 3): 'z', 4: g, '6': (7, 8)}, 'b': [1, 2, 3]}, 'z': -1}
    )
    self.assertEqual(
      repr(c),
      """Cursor(
  _obj={'a': {1: {(2, 3): 'z', 4: GenericTuple(x=1, y='a', z=(2, 'b')), '6': (7, 8)}, 'b': [1, 2, 3]}, 'z': -1},
  _changes={}
)""",
    )

    # test overwriting
    c['z'] = -2
    c['z'] = -3
    c['a']['b'][1] = -2
    c['a']['b'] = None

    # test deep mutation
    c['a'][1][4].x = (2, 4, 6)
    c['a'][1][4].z[0] = flax.core.freeze({'a': 1, 'b': {'c': 2, 'd': 3}})

    self.assertEqual(
      repr(c),
      """Cursor(
  _obj={'a': {1: {(2, 3): 'z', 4: GenericTuple(x=1, y='a', z=(2, 'b')), '6': (7, 8)}, 'b': [1, 2, 3]}, 'z': -1},
  _changes={
    'z': Cursor(
           _obj=-3,
           _changes={}
         ),
    'a': Cursor(
           _obj={1: {(2, 3): 'z', 4: GenericTuple(x=1, y='a', z=(2, 'b')), '6': (7, 8)}, 'b': [1, 2, 3]},
           _changes={
             'b': Cursor(
                    _obj=None,
                    _changes={}
                  ),
             1: Cursor(
                  _obj={(2, 3): 'z', 4: GenericTuple(x=1, y='a', z=(2, 'b')), '6': (7, 8)},
                  _changes={
                    4: Cursor(
                         _obj=GenericTuple(x=1, y='a', z=(2, 'b')),
                         _changes={
                           'x': Cursor(
                                  _obj=(2, 4, 6),
                                  _changes={}
                                ),
                           'z': Cursor(
                                  _obj=(2, 'b'),
                                  _changes={
                                    0: Cursor(
                                         _obj=FrozenDict({
                                             a: 1,
                                             b: {
                                                 c: 2,
                                                 d: 3,
                                             },
                                         }),
                                         _changes={}
                                       )
                                  }
                                )
                         }
                       )
                  }
                )
           }
         )
  }
)""",
    )

  def test_magic_methods(self):
    def same_value(v1, v2):
      if isinstance(v1, tuple):
        return all([
            jnp.all(jax.tree_util.tree_map(lambda x, y: x == y, e1, e2))
            for e1, e2 in zip(v1, v2)
        ])
      return jnp.all(jax.tree_util.tree_map(lambda x, y: x == y, v1, v2))

    list_obj = [(1, 2), (3, 4)]
    for l, tuple_wrap in ((list_obj, lambda x: x), (tuple(list_obj), tuple)):
      c = cursor(l)
      # test __len__
      self.assertTrue(same_value(len(c), len(l)))
      # test __iter__
      for i, child_c in enumerate(c):
        child_c[1] += i + 1
      self.assertEqual(c.build(), tuple_wrap([(1, 3), (3, 6)]))
      # test __reversed__
      for i, child_c in enumerate(reversed(c)):
        child_c[1] += i + 1
      self.assertEqual(c.build(), tuple_wrap([(1, 5), (3, 7)]))
    # test __iter__ error
    with self.assertRaisesRegex(
      NotImplementedError,
      '__iter__ method only implemented for tuples and lists, not type <class'
      " 'dict'>",
    ):
      c = cursor({'a': 1, 'b': 2})
      for key in c:
        c[key] *= -1
    # test __iter__ error
    with self.assertRaisesRegex(
      NotImplementedError,
      '__reversed__ method only implemented for tuples and lists, not type'
      " <class 'dict'>",
    ):
      c = cursor({'a': 1, 'b': 2})
      for key in reversed(c):
        c[key] *= -1

    for obj_value in (2, jnp.array([[1, -2], [3, 4]])):
      for c in (
        cursor(obj_value),
        cursor([obj_value])[0],
        cursor((obj_value,))[0],
        cursor({0: obj_value})[0],
        cursor(flax.core.freeze({0: obj_value}))[0],
        cursor(GenericTuple(x=obj_value)).x,
        cursor(GenericDataClass(x=obj_value)).x,
      ):
        # test __neg__
        self.assertTrue(same_value(-c, -obj_value))
        # test __pos__
        self.assertTrue(same_value(+c, +obj_value))
        # test __abs__
        self.assertTrue(same_value(abs(-c), abs(-obj_value)))
        # test __invert__
        self.assertTrue(same_value(~c, ~obj_value))
        # test __round__
        self.assertTrue(same_value(round(c + 0.123), round(obj_value + 0.123)))
        self.assertTrue(
          same_value(round(c + 0.123, 2), round(obj_value + 0.123, 2))
        )

        for other_value in (3, jnp.array([[5, 6], [7, 8]])):
          # test __add__
          self.assertTrue(same_value(c + other_value, obj_value + other_value))
          # test __radd__
          self.assertTrue(same_value(other_value + c, other_value + obj_value))
          # test __sub__
          self.assertTrue(same_value(c - other_value, obj_value - other_value))
          # test __rsub__
          self.assertTrue(same_value(other_value - c, other_value - obj_value))
          # test __mul__
          self.assertTrue(same_value(c * other_value, obj_value * other_value))
          # test __rmul__
          self.assertTrue(same_value(other_value * c, other_value * obj_value))
          # test __truediv__
          self.assertTrue(same_value(c / other_value, obj_value / other_value))
          # test __rtruediv__
          self.assertTrue(same_value(other_value / c, other_value / obj_value))
          # test __floordiv__
          self.assertTrue(
            same_value(c // other_value, obj_value // other_value)
          )
          # test __rfloordiv__
          self.assertTrue(
            same_value(other_value // c, other_value // obj_value)
          )
          # test __mod__
          self.assertTrue(same_value(c % other_value, obj_value % other_value))
          # test __rmod__
          self.assertTrue(same_value(other_value % c, other_value % obj_value))
          # test __divmod__
          self.assertTrue(
            same_value(divmod(c, other_value), divmod(obj_value, other_value))
          )
          # test __rdivmod__
          self.assertTrue(
            same_value(divmod(other_value, c), divmod(other_value, obj_value))
          )
          # test __pow__
          self.assertTrue(
            same_value(pow(c, other_value), pow(obj_value, other_value))
          )
          # test __rpow__
          self.assertTrue(
            same_value(pow(other_value, c), pow(other_value, obj_value))
          )
          # test __lshift__
          self.assertTrue(
            same_value(c << other_value, obj_value << other_value)
          )
          # test __rlshift__
          self.assertTrue(
            same_value(other_value << c, other_value << obj_value)
          )
          # test __rshift__
          self.assertTrue(
            same_value(c >> other_value, obj_value >> other_value)
          )
          # test __rrshift__
          self.assertTrue(
            same_value(other_value >> c, other_value >> obj_value)
          )
          # test __and__
          self.assertTrue(same_value(c & other_value, obj_value & other_value))
          # test __rand__
          self.assertTrue(same_value(other_value & c, other_value & obj_value))
          # test __xor__
          self.assertTrue(same_value(c ^ other_value, obj_value ^ other_value))
          # test __rxor__
          self.assertTrue(same_value(other_value ^ c, other_value ^ obj_value))
          # test __or__
          self.assertTrue(same_value(c | other_value, obj_value | other_value))
          # test __ror__
          self.assertTrue(same_value(other_value | c, other_value | obj_value))

          if isinstance(obj_value, jax.Array) and isinstance(
            other_value, jax.Array
          ):
            # test __matmul__
            self.assertTrue(
              same_value(c @ other_value, obj_value @ other_value)
            )
            # test __rmatmul__
            self.assertTrue(
              same_value(other_value @ c, other_value @ obj_value)
            )

          # test __lt__
          self.assertTrue(same_value(c < other_value, obj_value < other_value))
          self.assertTrue(same_value(other_value < c, other_value < obj_value))
          # test __le__
          self.assertTrue(
            same_value(c <= other_value, obj_value <= other_value)
          )
          self.assertTrue(
            same_value(other_value <= c, other_value <= obj_value)
          )
          # test __eq__
          self.assertTrue(
            same_value(c == other_value, obj_value == other_value)
          )
          self.assertTrue(
            same_value(other_value == c, other_value == obj_value)
          )
          # test __ne__
          self.assertTrue(
            same_value(c != other_value, obj_value != other_value)
          )
          self.assertTrue(
            same_value(other_value != c, other_value != obj_value)
          )
          # test __gt__
          self.assertTrue(same_value(c > other_value, obj_value > other_value))
          self.assertTrue(same_value(other_value > c, other_value > obj_value))
          # test __ge__
          self.assertTrue(
            same_value(c >= other_value, obj_value >= other_value)
          )
          self.assertTrue(
            same_value(other_value >= c, other_value >= obj_value)
          )

  def test_path(self):
    c = cursor(
      GenericTuple(
        x=[
          0,
          {'a': 1, 'b': (2, 3), ('c', 'd'): [4, 5]},
          (100, 200),
          [3, 4, 5],
        ],
        y=train_state.TrainState.create(
          apply_fn=lambda x: x,
          params=freeze({'a': 1, 'b': (2, 3), 'c': [4, 5]}),
          tx=optax.adam(1e-3),
        ),
      )
    )
    self.assertEqual(c.x[1][('c', 'd')][0]._path, ".x[1][('c', 'd')][0]")
    self.assertEqual(c.x[2][1]._path, '.x[2][1]')
    self.assertEqual(c.y.params['b'][1]._path, ".y.params['b'][1]")

    # test path when first access type is item access
    c = cursor([1, GenericTuple('a', 2), (3, 4)])
    self.assertEqual(c[1].x._path, '[1].x')
    self.assertEqual(c[2][0]._path, '[2][0]')

  def test_traverse_tree(self):
    c = cursor(
      GenericTuple(
        x=[
          0,
          {'a': 1, 'b': (2, 3), ('c', 'd'): [4, 5]},
          (100, 200),
          [3, 4, 5],
        ],
        y=3,
      )
    )

    def update_fn(path, value):
      if value == 4:
        return -4
      return value

    def cond_fn(path, value):
      return value == 3

    with self.assertRaisesRegex(
      TraverseTreeError,
      'Both update_fn and cond_fn are None. Exactly one of them must be'
      ' None.',
    ):
      next(_traverse_tree((), c._obj))
    with self.assertRaisesRegex(
      TraverseTreeError,
      'Both update_fn and cond_fn are not None. Exactly one of them must be'
      ' not None.',
    ):
      next(_traverse_tree((), c._obj, update_fn=update_fn, cond_fn=cond_fn))

    (p, v), (p2, v2) = _traverse_tree((), c._obj, update_fn=update_fn)
    self.assertEqual(
      p,
      (
        ('x', AccessType.ATTR),
        (1, AccessType.ITEM),
        (('c', 'd'), AccessType.ITEM),
        (0, AccessType.ITEM),
      ),
    )
    self.assertEqual(v, -4)
    self.assertEqual(
      p2, (('x', AccessType.ATTR), (3, AccessType.ITEM), (1, AccessType.ITEM))
    )
    self.assertEqual(v2, -4)

    p, p2, p3 = _traverse_tree((), c._obj, cond_fn=cond_fn)
    self.assertEqual(
      p,
      (
        ('x', AccessType.ATTR),
        (1, AccessType.ITEM),
        ('b', AccessType.ITEM),
        (1, AccessType.ITEM),
      ),
    )
    self.assertEqual(
      p2, (('x', AccessType.ATTR), (3, AccessType.ITEM), (0, AccessType.ITEM))
    )
    self.assertEqual(p3, (('y', AccessType.ATTR),))

  def test_set_and_build(self):
    # test regular dict and FrozenDict
    dict_obj = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    for d, freeze_wrap in ((dict_obj, lambda x: x), (freeze(dict_obj), freeze)):
      # set API
      self.assertEqual(
        cursor(d)['b'][0].set(10),
        freeze_wrap({'a': 1, 'b': (10, 3), 'c': [4, 5]}),
      )
      # build API
      c = cursor(d)
      c['b'][0] = 20
      c['a'] = (100, 200)
      d2 = c.build()
      self.assertEqual(
        d2, freeze_wrap({'a': (100, 200), 'b': (20, 3), 'c': [4, 5]})
      )
    self.assertEqual(
      dict_obj, {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    )  # make sure original object is unchanged

    # test list and tuple
    list_obj = [0, dict_obj, (1, 2), [3, 4, 5]]
    for l, tuple_wrap in ((list_obj, lambda x: x), (tuple(list_obj), tuple)):
      # set API
      self.assertEqual(
        cursor(l)[1]['b'][0].set(10),
        tuple_wrap([0, {'a': 1, 'b': (10, 3), 'c': [4, 5]}, (1, 2), [3, 4, 5]]),
      )
      # build API
      c = cursor(l)
      c[1]['b'][0] = 20
      c[2] = (100, 200)
      l2 = c.build()
      self.assertEqual(
        l2,
        tuple_wrap(
          [0, {'a': 1, 'b': (20, 3), 'c': [4, 5]}, (100, 200), [3, 4, 5]]
        ),
      )
    self.assertEqual(
      list_obj, [0, {'a': 1, 'b': (2, 3), 'c': [4, 5]}, (1, 2), [3, 4, 5]]
    )  # make sure original object is unchanged

    # test TrainState
    state = train_state.TrainState.create(
      apply_fn=lambda x: x,
      params=dict_obj,
      tx=optax.adam(1e-3),
    )
    # set API
    self.assertEqual(
      cursor(state).params['b'][0].set(10).params,
      {'a': 1, 'b': (10, 3), 'c': [4, 5]},
    )
    # build API
    new_fn = lambda x: x + 1
    c = cursor(state)
    c.apply_fn = new_fn
    c.params['b'][0] = 20
    c.params['a'] = (100, 200)
    state2 = c.build()
    self.assertEqual(state2.apply_fn, new_fn)
    self.assertEqual(
      state2.params, {'a': (100, 200), 'b': (20, 3), 'c': [4, 5]}
    )

    self.assertEqual(
      dict_obj, {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    )  # make sure original object is unchanged

    # test NamedTuple
    # set API
    t = GenericTuple(GenericTuple(0))
    self.assertEqual(cursor(t).x.x.set(1), GenericTuple(GenericTuple(1)))

    # build API
    c = cursor(t)
    c.x.x = 2
    c.x.y = 3
    c.y = 4
    t2 = c.build()
    self.assertEqual(t2, GenericTuple(GenericTuple(2, 3), 4))

    self.assertEqual(
      t, GenericTuple(GenericTuple(0))
    )  # make sure original object is unchanged

  def test_apply_update(self):
    # test list and tuple
    def update_fn(path, value):
      """Multiply the first element of all leaf nodes of the pytree by -1."""
      if path[-1] == '0' and isinstance(value, int):
        return value * -1
      return value

    for tuple_wrap in (lambda x: x, tuple):
      l = tuple_wrap([tuple_wrap([1, 2]), tuple_wrap([3, 4])])
      c = cursor(l)
      l2 = c.apply_update(update_fn).build()
      self.assertEqual(
        l2, tuple_wrap([tuple_wrap([-1, 2]), tuple_wrap([-3, 4])])
      )
      self.assertEqual(
        l, tuple_wrap([tuple_wrap([1, 2]), tuple_wrap([3, 4])])
      )  # make sure the original object is unchanged

    # test regular dict and FrozenDict
    def update_fn(path, value):
      """Multiply all dense kernel params by 2 and add 1.
      Subtract the Dense_1 bias param by 1."""
      if 'kernel' in path:
        return value * 2 + 1
      elif 'Dense_1' in path and 'bias' in path:
        return value - 1
      return value

    class Model(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        return x

    for freeze_wrap in (lambda x: x, freeze):
      params = freeze_wrap(
        Model().init(jax.random.key(0), jnp.empty((1, 2)))['params']
      )

      c = cursor(params)
      params2 = c.apply_update(update_fn).build()
      for layer in ('Dense_0', 'Dense_1', 'Dense_2'):
        self.assertTrue(
          (params2[layer]['kernel'] == 2 * params[layer]['kernel'] + 1).all()
        )
        if layer == 'Dense_1':
          self.assertTrue(
            (params2[layer]['bias'] == jnp.array([-1, -1, -1])).all()
          )
        else:
          self.assertTrue(
            (params2[layer]['bias'] == params[layer]['bias']).all()
          )
      self.assertTrue(
        jax.tree_util.tree_all(
          jax.tree_util.tree_map(
            lambda x, y: (x == y).all(),
            params,
            freeze_wrap(
              Model().init(jax.random.key(0), jnp.empty((1, 2)))['params']
            ),
          )
        )
      )  # make sure original params are unchanged

    # test TrainState
    def update_fn(path, value):
      """Replace params with empty dictionary."""
      if 'params' in path:
        return {}
      return value

    state = train_state.TrainState.create(
      apply_fn=lambda x: x,
      params={'a': 1, 'b': 2},
      tx=optax.adam(1e-3),
    )
    c = cursor(state)
    state2 = c.apply_update(update_fn).build()
    self.assertEqual(state2.params, {})
    self.assertEqual(
      state.params, {'a': 1, 'b': 2}
    )  # make sure original params are unchanged

    # test NamedTuple
    def update_fn(path, value):
      """Add 5 to all x-attribute values that are ints."""
      if path[-1] == 'x' and isinstance(value, int):
        return value + 5
      return value

    t = GenericTuple(GenericTuple(0, 1), GenericTuple(2, 3))
    c = cursor(t)
    t2 = c.apply_update(update_fn).build()
    self.assertEqual(t2, GenericTuple(GenericTuple(5, 1), GenericTuple(7, 3)))
    self.assertEqual(
      t, GenericTuple(GenericTuple(0, 1), GenericTuple(2, 3))
    )  # make sure original object is unchanged

  def test_apply_update_root_node_unmodified(self):
    def update_fn(path, value):
      if isinstance(value, list):
        value = value.copy()
        value.append(-1)
      return value

    l = [[1, 2], [3, 4], 5]
    l2 = cursor(l).apply_update(update_fn).build()
    self.assertEqual(l2, [[1, 2, -1], [3, 4, -1], 5])

  def test_multi_modify(self):
    d = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    c = cursor(d)
    # test multiple changes on same element
    c['b'][0] = 6
    c['b'][0] = 7
    d2 = c.build()
    self.assertEqual(d2, {'a': 1, 'b': (7, 3), 'c': [4, 5]})
    # test nested changes
    c['a'] = (100, 200)
    c['a'][1] = -1
    d3 = c.build()
    self.assertEqual(d3, {'a': (100, -1), 'b': (7, 3), 'c': [4, 5]})

  def test_hidden_change(self):
    # test list
    l = [1, 2]
    c = cursor(l)
    c[0] = 100
    l[1] = -1
    l2 = c.build()
    self.assertEqual(
      l2, [100, -1]
    )  # change in l affects l2 (this is expected behavior)
    self.assertEqual(l, [1, -1])

    # test regular dict
    d = {'a': 1, 'b': 2}
    c = cursor(d)
    c['a'] = 100
    d['b'] = -1
    d2 = c.build()
    self.assertEqual(
      d2, {'a': 100, 'b': -1}
    )  # change in d affects d2 (this is expected behavior)
    self.assertEqual(d, {'a': 1, 'b': -1})

    # test TrainState
    params = {'a': 1, 'b': 2}
    state = train_state.TrainState.create(
      apply_fn=lambda x: x,
      params=params,
      tx=optax.adam(1e-3),
    )
    c = cursor(state)
    c.params['a'] = 100
    params['b'] = -1
    state2 = c.build()
    self.assertEqual(
      state2.params, {'a': 100, 'b': -1}
    )  # change in state affects state2 (this is expected behavior)
    self.assertEqual(state.params, {'a': 1, 'b': -1})

  def test_named_tuple_multi_access(self):
    t = GenericTuple(GenericTuple(0, 1), GenericTuple(2, 3))
    c = cursor(t)
    c.x.x = 4
    c[0].y = 5
    c.y.x = 6
    c.y[1] = 7
    self.assertEqual(
      c.build(), GenericTuple(GenericTuple(4, 5), GenericTuple(6, 7))
    )

    c[0][1] = -5
    self.assertEqual(
      c.build(), GenericTuple(GenericTuple(4, -5), GenericTuple(6, 7))
    )
    c.x[1] = -6
    self.assertEqual(
      c.build(), GenericTuple(GenericTuple(4, -6), GenericTuple(6, 7))
    )
    c.x.y = -7
    self.assertEqual(
      c.build(), GenericTuple(GenericTuple(4, -7), GenericTuple(6, 7))
    )
    c[0].y = -8
    self.assertEqual(
      c.build(), GenericTuple(GenericTuple(4, -8), GenericTuple(6, 7))
    )

  def test_find(self):
    c = cursor(
      GenericTuple(
        x=[
          0,
          {'a': 1, 'b': (2, 3), ('c', 'd'): [4, 5]},
          (100, 200),
          [3, 4, 5],
        ],
        y=train_state.TrainState.create(
          apply_fn=lambda x: x,
          params=freeze({'a': 1, 'b': (2, 3), 'c': [4, 5]}),
          tx=optax.adam(1e-3),
        ),
      )
    )

    with self.assertRaisesRegex(
      CursorFindError,
      'More than one object found given the conditions of the cond_fn\\. '
      'The first two objects found have the following paths: '
      "\\.x\\[1]\\['b'] and \\.y\\.params\\['b'] ",
    ):
      c.find(lambda path, value: 'b' in path and isinstance(value, tuple))
    with self.assertRaisesRegex(
      CursorFindError,
      'No object found given the conditions of the cond_fn\\.',
    ):
      c.find(lambda path, value: 'b' in path and isinstance(value, str))

    self.assertEqual(
      c.find(lambda path, value: path.endswith('params/b'))[1].set(30).y.params,
      freeze({'a': 1, 'b': (2, 30), 'c': [4, 5]}),
    )

  def test_find_all(self):
    # test list and tuple
    def cond_fn(path, value):
      """Get all lists that are not the first element in its parent."""
      return path[-1] != '0' and isinstance(value, (tuple, list))

    for tuple_wrap in (lambda x: x, tuple):
      l = tuple_wrap(
        [tuple_wrap([1, 2]), tuple_wrap([3, 4]), tuple_wrap([5, 6])]
      )
      c = cursor(l)
      c2, c3 = c.find_all(cond_fn)
      c2[0] *= -1
      c3[1] *= -2
      self.assertEqual(
        c.build(),
        tuple_wrap(
          [tuple_wrap([1, 2]), tuple_wrap([-3, 4]), tuple_wrap([5, -12])]
        ),
      )
      self.assertEqual(
        l,
        tuple_wrap(
          [tuple_wrap([1, 2]), tuple_wrap([3, 4]), tuple_wrap([5, 6])]
        ),
      )  # make sure the original object is unchanged

    # test regular dict and FrozenDict
    def cond_fn(path, value):
      """Get the second and third dense params."""
      return 'Dense_1' in path or 'Dense_2' in path

    class Model(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        x = nn.Dense(3)(x)
        x = nn.relu(x)
        return x

    for freeze_wrap in (lambda x: x, freeze):
      params = freeze_wrap(
        Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))['params']
      )
      c = cursor(params)
      for i, c2 in enumerate(c.find_all(cond_fn)):
        self.assertEqual(
          c2['kernel'].set(123)[f'Dense_{i+1}'],
          freeze_wrap({'kernel': 123, 'bias': params[f'Dense_{i+1}']['bias']}),
        )
      self.assertTrue(
        jax.tree_util.tree_all(
          jax.tree_util.tree_map(
            lambda x, y: (x == y).all(),
            params,
            freeze_wrap(
              Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))['params']
            ),
          )
        )
      )  # make sure original params are unchanged

    # test TrainState
    def cond_fn(path, value):
      """Find TrainState params."""
      return 'params' in path

    state = train_state.TrainState.create(
      apply_fn=lambda x: x,
      params={'a': 1, 'b': 2},
      tx=optax.adam(1e-3),
    )
    c = cursor(state)
    c2 = list(c.find_all(cond_fn))
    self.assertEqual(len(c2), 1)
    c2 = c2[0]
    self.assertEqual(c2['b'].set(-1).params, {'a': 1, 'b': -1})
    self.assertEqual(
      state.params, {'a': 1, 'b': 2}
    )  # make sure original params are unchanged

    # test NamedTuple
    def cond_fn(path, value):
      """Get all GenericTuples that have int x-attribute values."""
      return isinstance(value, GenericTuple) and isinstance(value.x, int)

    t = GenericTuple(
      GenericTuple(0, 'a'), GenericTuple(1, 'b'), GenericTuple('c', 2)
    )
    c = cursor(t)
    c2, c3 = c.find_all(cond_fn)
    c2.x += 5
    c3.x += 6
    self.assertEqual(
      c.build(),
      GenericTuple(
        GenericTuple(5, 'a'), GenericTuple(7, 'b'), GenericTuple('c', 2)
      ),
    )
    self.assertEqual(
      t,
      GenericTuple(
        GenericTuple(0, 'a'), GenericTuple(1, 'b'), GenericTuple('c', 2)
      ),
    )  # make sure original object is unchanged


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import jax.numpy as jnp
import optax

from flax import nnx
from flax.nnx.nnx import filterlib, graph
from flax.nnx.nnx.object import Object
from flax.nnx.nnx.variables import Variable

# TODO: add tests and docstrings


class OptState(Variable):
  """Wrapper class for Optimizer Variables."""

  pass


class Optimizer(Object):
  """Simple train state for the common case with a single Optax optimizer.

  Example usage::

    >>> import jax, jax.numpy as jnp
    >>> from flax import nnx
    >>> import optax
    ...
    >>> class Model(nnx.Module):
    ...   def __init__(self, rngs):
    ...     self.linear1 = nnx.Linear(2, 3, rngs=rngs)
    ...     self.linear2 = nnx.Linear(3, 4, rngs=rngs)
    ...   def __call__(self, x):
    ...     return self.linear2(self.linear1(x))
    ...
    >>> x = jax.random.normal(jax.random.key(0), (1, 2))
    >>> y = jnp.ones((1, 4))
    ...
    >>> model = Model(nnx.Rngs(0))
    >>> tx = optax.adam(1e-3)
    >>> state = nnx.Optimizer(model, tx)
    ...
    >>> loss_fn = lambda model: ((model(x) - y) ** 2).mean()
    >>> loss_fn(model)
    Array(1.7055722, dtype=float32)
    >>> grads = nnx.grad(loss_fn, wrt=nnx.Param)(state.model)
    >>> state.update(grads)
    >>> loss_fn(model)
    Array(1.6925814, dtype=float32)

  Note that you can easily extend this class by subclassing it for storing
  additional data (e.g. adding metrics).

  Example usage::

    >>> class TrainState(nnx.Optimizer):
    ...   def __init__(self, model, tx, metrics):
    ...     self.metrics = metrics
    ...     super().__init__(model, tx)
    ...   def update(self, *, grads, **updates):
    ...     self.metrics.update(**updates)
    ...     super().update(grads)
    ...
    >>> metrics = nnx.metrics.Average()
    >>> state = TrainState(model, tx, metrics)
    ...
    >>> grads = nnx.grad(loss_fn, wrt=nnx.Param)(state.model)
    >>> state.update(grads=grads, values=loss_fn(state.model))
    >>> state.metrics.compute()
    Array(1.6925814, dtype=float32)
    >>> state.update(grads=grads, values=loss_fn(state.model))
    >>> state.metrics.compute()
    Array(1.68612, dtype=float32)

  For more exotic usecases (e.g. multiple optimizers) it's probably best to
  fork the class and modify it.

  Attributes:
    step: An ``OptState`` :class:`Variable` that tracks the step count.
    model: The wrapped :class:`Module`.
    tx: An Optax gradient transformation.
    opt_state: The Optax optimizer state.
  """

  def __init__(
    self,
    model: nnx.Module,
    tx: optax.GradientTransformation,
    wrt: filterlib.Filter = nnx.Param,
  ):
    """
    Instantiate the class and wrap the :class:`Module` and Optax gradient
    transformation. Instantiate the optimizer state to keep track of
    :class:`Variable` types specified in ``wrt``. Set the step count to 0.

    Args:
      model: An NNX Module.
      tx: An Optax gradient transformation.
      wrt: optional argument to filter for which :class:`Variable`'s to keep
        track of in the optimizer state. These should be the :class:`Variable`'s
        that you plan on updating; i.e. this argument value should match the
        ``wrt``  argument passed to the ``nnx.grad`` call that will generate the
        gradients that will be passed into the ``grads`` argument of the
        :func:`update` method.
    """
    self.step = OptState(jnp.array(0, dtype=jnp.uint32))
    self.model = model
    self.tx = tx
    self.opt_state = tx.init(nnx.state(model, wrt))
    self.wrt = wrt

  def split(self, *filters: filterlib.Filter):
    return graph.split(self, *filters)

  def update(self, grads):
    """Updates ``step``, ``params``, ``opt_state`` and ``**kwargs`` in return value.
    The ``grads`` must be derived from ``nnx.grad(..., wrt=self.wrt)``, where the
    gradients are with respect to the same :class:`Variable` types as defined in
    ``self.wrt`` during instantiation of this ``Optimizer``. For example::

      >>> from flax import nnx
      >>> import jax, jax.numpy as jnp
      >>> import optax

      >>> class CustomVariable(nnx.Variable):
      ...   pass

      >>> class Model(nnx.Module):
      ...   def __init__(self, rngs):
      ...     self.linear = nnx.Linear(2, 3, rngs=rngs)
      ...     self.custom_variable = CustomVariable(jnp.ones((1, 3)))
      ...   def __call__(self, x):
      ...     return self.linear(x) + self.custom_variable
      >>> model = Model(rngs=nnx.Rngs(0))
      >>> jax.tree.map(jnp.shape, nnx.state(model))
      State({
        'custom_variable': VariableState(
          type=CustomVariable,
          value=(1, 3)
        ),
        'linear': {
          'bias': VariableState(
            type=Param,
            value=(3,)
          ),
          'kernel': VariableState(
            type=Param,
            value=(2, 3)
          )
        }
      })

      >>> # update:
      >>> # - only Linear layer parameters
      >>> # - only CustomVariable parameters
      >>> # - both Linear layer and CustomVariable parameters
      >>> loss_fn = lambda model, x, y: ((model(x) - y) ** 2).mean()
      >>> for variable in (nnx.Param, CustomVariable, (nnx.Param, CustomVariable)):
      ...   # make sure `wrt` arguments match for `nnx.Optimizer` and `nnx.grad`
      ...   state = nnx.Optimizer(model, optax.adam(1e-3), wrt=variable)
      ...   grads = nnx.grad(loss_fn, wrt=variable)(
      ...     state.model, jnp.ones((1, 2)), jnp.ones((1, 3))
      ...   )
      ...   state.update(grads=grads)

    Note that internally this function calls ``.tx.update()`` followed by a call
    to ``optax.apply_updates()`` to update ``params`` and ``opt_state``.

    Args:
      grads: the gradients derived from ``nnx.grad``.
    """
    state = nnx.state(self.model, self.wrt)

    updates, new_opt_state = self.tx.update(grads, self.opt_state, state)
    new_params = optax.apply_updates(state, updates)
    assert isinstance(new_params, nnx.State)

    self.step.value += 1
    nnx.update(self.model, new_params)
    self.opt_state = new_opt_state



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copied over from MaxText (https://github.com/google/maxtext/blob/main/MaxText/max_utils.py).

import functools
import logging

import numpy as np
import flax.linen as nn
from flax.linen import partitioning as nn_partitioning
from flax.training import train_state
import jax
import jax.numpy as jnp
from jax.experimental import mesh_utils


# Mesh utils.
# -----------------------------------------------------------------------------


def create_device_mesh(config):
  """Creates a device mesh with each slice in its own data parallel group. If there is only one slice, uses two replicas."""
  devices = jax.devices()
  num_devices = len(devices)
  try:
    num_slices = 1 + max([d.slice_index for d in devices])
  except:
    num_slices = 1
  num_devices_per_slice = num_devices // num_slices
  logging.info(f"Devices: {devices}")
  logging.info(f"Number of devices: {num_devices}")

  multi_slice_env = hasattr(jax.devices()[0], "slice_index")

  dcn_parallelism = [
      config.dcn_data_parallelism,
      config.dcn_fsdp_parallelism,
      config.dcn_tensor_parallelism,
  ]
  ici_parallelism = [
      config.ici_data_parallelism,
      config.ici_fsdp_parallelism,
      config.ici_tensor_parallelism,
  ]

  # Find possible unspecified parallelisms
  dcn_parallelism = fill_unspecified_mesh_axes(
      dcn_parallelism, num_slices, "DCN"
  )
  ici_parallelism = fill_unspecified_mesh_axes(
      ici_parallelism, num_devices_per_slice, "ICI"
  )

  if multi_slice_env:
    mesh = mesh_utils.create_hybrid_device_mesh(
        ici_parallelism, dcn_parallelism
    )
  else:
    mesh = mesh_utils.create_device_mesh(ici_parallelism)

  logging.info(f"Decided on mesh: {mesh}")
  logging.info(f"Mesh shape: {mesh.shape}")

  return mesh


def fill_unspecified_mesh_axes(
    parallelism_vals, target_product, parallelism_type
):
  """Evaluates unspecified DCN/ICI parallelism values"""
  if -1 in parallelism_vals:
    assert parallelism_vals.count(-1) == 1, (
        f"Found unspecified values (-1) for more than one {parallelism_type}   "
        "   parallelism axis. At most one axis can be unspecified."
    )

    determined_val = target_product / np.prod(parallelism_vals) * -1

    assert determined_val >= 1 and determined_val.is_integer, (
        "Unspecified value unable to be determined with the given     "
        f" {parallelism_type} parallelism values"
    )

    parallelism_vals[parallelism_vals.index(-1)] = int(determined_val)

  target_type = "slices" if parallelism_type == "DCN" else "devices per slice"

  assert np.prod(parallelism_vals) == target_product, (
      f"Number of {target_type} {target_product} does not match    the product"
      f" of the {parallelism_type} parallelism {np.prod(parallelism_vals)}"
  )

  return parallelism_vals


# State initialization utils.
# -----------------------------------------------------------------------------


def unbox_logicallypartioned_trainstate(
    boxed_train_state: train_state.TrainState,
):
  """Unboxes the flax.LogicallyPartitioned pieces in a train state.

  Args:
    boxed_train_state: a train state that includes LogicallyPartitioned
      leaves.
  Returns:
    a TrainState where all all LogicallyPartitioned leaves have been unboxed.
  """
  return jax.tree_util.tree_map(
      lambda x: x.unbox() if isinstance(x, nn.spmd.LogicallyPartitioned) else x,
      boxed_train_state,
      is_leaf=lambda k: isinstance(k, nn.spmd.LogicallyPartitioned),
  )


def init_train_state(model, tx, config, key):
  """
  We pass in "static" objects like model, tx, config as JAX compares them by
  object hash, and instantiating them inside causes pjit top-level annotations
  to fail to match as pytree prefixes if we re-instantiate.

  Args: model, tx, config, key
  """
  input_shape = (config.per_device_batch_size, config.max_target_length)
  initial_variables = jax.jit(model.init)(
      key, jnp.ones(input_shape, jnp.float32)
  )

  state = train_state.TrainState.create(
      apply_fn=model.apply, params=initial_variables["params"], tx=tx
  )
  return state


def setup_initial_state(model, tx, config, rng, mesh):
  """We initialize the model and optimizer state, and optionally load from a
  checkpoint as necessary.

  Args:
    model: the flax model to initialize
    tx: the optax.GradientTransformation
    config: config object
    rng: jax.prng key
    mesh: jax.devices() mesh

  Returns:
    state: the initialized train state
    state_mesh_annotations: the mesh annotations for the train state
  """
  init_train_state_partial = functools.partial(
      init_train_state, model, tx, config
  )
  abstract_state = jax.eval_shape(init_train_state_partial, rng)
  state_logical_annotations = nn.get_partition_spec(abstract_state)

  # Initialization
  with mesh, nn_partitioning.axis_rules(config.logical_axis_rules):
    state_mesh_annotations = nn.logical_to_mesh_sharding(
        state_logical_annotations, mesh, config.logical_axis_rules
    )
    state = jax.jit(
        init_train_state_partial,
        in_shardings=None,  # type: ignore
        out_shardings=state_mesh_annotations,
    )(rng)

  state = unbox_logicallypartioned_trainstate(state)
  return state, state_mesh_annotations



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility script for updating examples' "requirements.txt".

WARNING: Updates examples/*/requirements.txt in place, so you probably want to
run this in a clean working directory...

This script is useful to automatically update the requirements of the individual
examples. It's desirable to have these dependencies fully specified and
guaranteed to be complete in order to download an example directly from Github
and run it on a virtual machine without any additional manual package
installation (note though that jax/jaxlib must already be installed).

To update all "requirements.txt" the program is run with a `--versions` argument
that specifies the version of every package that is required by any of the
examples. If any of the versions is not specified, the program will fail with an
error.

Example invocations (from the Flax repo root directory):

python dev --versions='absl-py-0.12.0 clu-0.0.6 flax-0.3.6'

The actual pypi-version list can be read copied from the "Install Dependencies"
step from the latest Github build action:
https://github.com/google/flax/actions/workflows/build.yml

Alternatively, the list can also be provided from the local environment with:

python dev --versions="$(pip freeze | sed s/==/-/g) flax-0.3.6"

This will use the currently installed versions of all packages.
"""

import pathlib
import re

from absl import app, flags

FLAGS = flags.FLAGS

flags.DEFINE_string(
  'versions',
  None,
  'space-separated list of pkg-name-1.0.0 pairs. can be obtained from local '
  'environment with '
  '`--version="$(pip freeze | sed s/==/-/g) flax-0.3.6"` '
  '(note the flax version "override") '
  'or from the "install dependencies" step in the github build action '
  'https://github.com/google/flax/actions/workflows/build.yml',
)
flags.mark_flag_as_required('versions')
flags.DEFINE_bool('verbose', False, 'enables verbose output.')
flags.DEFINE_list('ignore', ['jax'], 'packages not to add to requirements.')


import_re = re.compile(r'(?:from|import)\s+(\w+)')
# maps `import cv2` to `pip install opencv-python`
pkg_map = {
  'absl': 'absl-py',
  'atari_py': 'atari-py',
  'cv2': 'opencv-python',
  'ml_collections': 'ml-collections',
  'PIL': 'Pillow',
  'tensorflow_datasets': 'tensorflow-datasets',
  'tensorflow_text': 'tensorflow-text',
}
standard_libs = set(
  'codecs collections dataclasses datetime enum functools math'
  ' multiprocessing itertools os pathlib random re sys tempfile time typing'
  ' unicodedata warnings'.split(' ')
)


def main(argv):
  del argv

  versions = {
    pkg_version[: pkg_version.rindex('-')]: pkg_version[
      pkg_version.rindex('-') + 1 :
    ]
    for pkg_version in FLAGS.versions.replace('\n', ' ').split(' ')
    if '-' in pkg_version
  }
  if FLAGS.verbose:
    print('parsed versions:', sorted(versions.items()))
  ignore = set(FLAGS.ignore)

  examples_root = pathlib.Path(__file__).absolute().parents[1] / 'examples'

  for example_dir in examples_root.glob('*'):
    if not example_dir.is_dir():
      continue
    requirements = example_dir / 'requirements.txt'
    if not requirements.exists():
      print(f'Skipping directory "{example_dir}"')
      continue

    pkgs = set()
    local_pkgs_and_modules = set()
    for path in example_dir.glob('*'):
      if path.is_dir() and any(path.glob('*.py')):
        local_pkgs_and_modules.add(path.parts[-1])  # local package
    for py in example_dir.glob('*.py'):
      local_pkgs_and_modules.add(py.parts[-1][:-3])  # local module
      for line in open(py):
        m = import_re.match(line)
        if m:
          pkgs.add(m.group(1))

    pkgs -= local_pkgs_and_modules | standard_libs

    print(f'{requirements} -', end=' ')
    with requirements.open('w') as f:
      for pkg in sorted(pkgs, key=str.casefold):
        if pkg in ignore:
          continue
        pkg = pkg_map.get(pkg, pkg)
        print(f'{pkg}-{versions[pkg]}', end=' ')
        f.write(f'{pkg}=={versions[pkg]}\n')
      print()


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for helpers."""

from __future__ import annotations

from typing import Tuple

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
import flax.linen as nn
import helpers
import jax
import jax.numpy as jnp
import numpy as np


class ModuleFromLinenVariablesTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          inputs_shape=(1, 2, 3, 4),
          num_features=10,
          use_bias=True,
      ),
      dict(
          inputs_shape=(10, 5),
          num_features=4,
          use_bias=False,
      ),
  )
  def test_same_structure(self, inputs_shape, num_features, use_bias):
    rng_key = jax.random.PRNGKey(0)
    rng_inputs, rng_params = jax.random.split(rng_key)
    inputs = jax.random.normal(rng_inputs, inputs_shape)

    linen_mdl = nn.Dense(
        features=num_features,
        use_bias=use_bias,
    )
    linen_init_vars = linen_mdl.init(rng_params, jnp.zeros(inputs_shape))
    linen_output = linen_mdl.apply(
        linen_init_vars,
        inputs,
    )

    mdl = helpers.module_from_linen_variables(
        module_factory=lambda: nnx.Linear(
            in_features=inputs_shape[-1],
            out_features=num_features,
            use_bias=use_bias,
            rngs=nnx.Rngs(params=rng_params),
        ),
        variables=linen_init_vars,
    )
    output = mdl(inputs)

    np.testing.assert_array_equal(output, linen_output)

  @parameterized.parameters(
      dict(
          inputs_shape=(1, 2, 3, 4),
          num_features=(10, 20, 7),
          use_bias=(False, True, False),
      ),
  )
  def test_different_structure(self, inputs_shape, num_features, use_bias):
    rng_key = jax.random.PRNGKey(0)
    rng_inputs, rng_params = jax.random.split(rng_key)
    inputs = jax.random.normal(rng_inputs, inputs_shape)

    linen_mdl = nn.Sequential([
        nn.Sequential([
            nn.BatchNorm(use_running_average=False),
            nn.Dense(features=f, use_bias=b),
        ])
        for f, b in zip(num_features, use_bias)
    ])
    linen_init_vars = linen_mdl.init(rng_key, jnp.zeros(inputs_shape))
    linen_output, linen_vars = linen_mdl.apply(
        linen_init_vars,
        inputs,
        mutable=['batch_stats'],
    )

    module_factory = lambda: nnx.Sequential(*[
        nnx.Sequential(
            nnx.BatchNorm(
                num_features=in_f,
                use_running_average=False,
                rngs=nnx.Rngs(params=rng_params),
            ),
            nnx.Linear(
                in_features=in_f,
                out_features=out_f,
                use_bias=b,
                rngs=nnx.Rngs(params=rng_params),
            ),
        )
        for in_f, out_f, b in zip(in_features, out_features, use_bias)
    ])

    def _map_key_fn(key: Tuple[str, ...]) -> Tuple[str | int, ...]:
      new_key = []
      for k in key[1:]:
        if k.startswith('layers_'):
          prefix, suffix = k.split('layers_')
          assert not prefix, prefix
          new_key.append('layers')
          new_key.append(int(suffix))
        else:
          new_key.append(k)

      return tuple(new_key)

    in_features = (inputs_shape[-1], *num_features[:-1])
    out_features = num_features
    mdl = helpers.module_from_linen_variables(
        module_factory=module_factory,
        variables=linen_init_vars,
        map_key_fn=_map_key_fn,
    )
    output = mdl(inputs)

    np.testing.assert_array_equal(output, linen_output)
    for i in range(len(num_features)):
      np.testing.assert_array_equal(
          mdl.layers[i].layers[0].mean.value,
          linen_vars['batch_stats'][f'layers_{i}']['layers_0']['mean'],
      )
      np.testing.assert_array_equal(
          mdl.layers[i].layers[0].var.value,
          linen_vars['batch_stats'][f'layers_{i}']['layers_0']['var'],
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is created with reference to torchvision/utils.py.

Modify: torch.tensor -> jax.numpy.DeviceArray
If you want to know about this file in detail, please visit the original code:
    https://github.com/pytorch/vision/blob/master/torchvision/utils.py
"""
import math

import jax.numpy as jnp
import numpy as np
from PIL import Image


def save_image(ndarray, fp, nrow=8, padding=2, pad_value=0.0, format_img=None):
  """Make a grid of images and Save it into an image file.

  Args:
    ndarray (array_like): 4D mini-batch images of shape (B x H x W x C)
    fp:  A filename(string) or file object
    nrow (int, optional): Number of images displayed in each row of the grid.
      The final grid size is ``(B / nrow, nrow)``. Default: ``8``.
    padding (int, optional): amount of padding. Default: ``2``.
    pad_value (float, optional): Value for the padded pixels. Default: ``0``.
    format_img(Optional):  If omitted, the format to use is determined from the
      filename extension. If a file object was used instead of a filename,
      this parameter should always be used.
  """

  if not (
      isinstance(ndarray, jnp.ndarray)
      or (
          isinstance(ndarray, list)
          and all(isinstance(t, jnp.ndarray) for t in ndarray)
      )
  ):
    raise TypeError(f'array_like of tensors expected, got {type(ndarray)}')

  ndarray = jnp.asarray(ndarray)

  if ndarray.ndim == 4 and ndarray.shape[-1] == 1:  # single-channel images
    ndarray = jnp.concatenate((ndarray, ndarray, ndarray), -1)

  # make the mini-batch of images into a grid
  nmaps = ndarray.shape[0]
  xmaps = min(nrow, nmaps)
  ymaps = int(math.ceil(float(nmaps) / xmaps))
  height, width = (
      int(ndarray.shape[1] + padding),
      int(ndarray.shape[2] + padding),
  )
  num_channels = ndarray.shape[3]
  grid = jnp.full(
      (height * ymaps + padding, width * xmaps + padding, num_channels),
      pad_value,
  ).astype(jnp.float32)
  k = 0
  for y in range(ymaps):
    for x in range(xmaps):
      if k >= nmaps:
        break
      grid = grid.at[
          y * height + padding : (y + 1) * height,
          x * width + padding : (x + 1) * width,
      ].set(ndarray[k])
      k = k + 1

  # Add 0.5 after unnormalizing to [0, 255] to round to nearest integer
  ndarr = np.array(jnp.clip(grid * 255.0 + 0.5, 0, 255).astype(jnp.uint8))
  im = Image.fromarray(ndarr.copy())
  im.save(fp, format=format_img)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random, sharding
from jax.experimental import mesh_utils

from flax import errors
from flax.core import init, lift, meta, nn


class MetaTest(absltest.TestCase):
  def test_boxed_param(self):
    def f(scope, xs):
      def g(scope, x):
        kernel_init = meta.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        kernel = scope.param('kernel', kernel_init, (x.shape[-1], 2))
        kernel_box = scope.get_variable('params', 'kernel')
        self.assertIsInstance(kernel_box, meta.Partitioned)
        self.assertEqual(kernel_box.names, ('in', 'out'))
        return x @ kernel

      lift.vmap(
        g,
        in_axes=0,
        variable_axes={'params': 0},
        split_rngs={'params': True},
        metadata_params={meta.PARTITION_NAME: 'batch'},
      )(scope, xs)

    _, variables = init(f)(random.key(0), jnp.zeros((8, 3)))
    self.assertEqual(
      variables['params']['kernel'].names, ('batch', 'in', 'out')
    )

  def test_boxed_variable(self):
    def f(scope, xs):
      def g(scope, x):
        kernel_init = meta.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        kernel = scope.variable(
          'params',
          'kernel',
          kernel_init,
          scope.make_rng('params'),
          (x.shape[-1], 2),
        )
        kernel.value += 1.0
        self.assertEqual(kernel.value.sum(), kernel.value.size * 2)
        kernel_box = scope.get_variable('params', 'kernel')
        self.assertIsInstance(kernel_box, meta.Partitioned)
        self.assertEqual(kernel_box.names, ('in', 'out'))
        return x @ kernel.value

      lift.vmap(
        g,
        in_axes=0,
        variable_axes={'params': 0},
        split_rngs={'params': True},
        metadata_params={meta.PARTITION_NAME: 'batch'},
      )(scope, xs)

    _, variables = init(f)(random.key(0), jnp.zeros((8, 3)))
    self.assertEqual(
      variables['params']['kernel'].names, ('batch', 'in', 'out')
    )

  def test_partition_axis_unspecified(self):
    def f(scope, xs):
      def g(scope, x):
        kernel_init = meta.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        scope.param('kernel', kernel_init, (3, 2))
        return x

      with self.assertRaises(errors.PartitioningUnspecifiedError):
        lift.vmap(
          g,
          in_axes=0,
          variable_axes={'params': 0},
          split_rngs={'params': True},
          metadata_params={},
        )(scope, xs)

    init(f)(random.key(0), jnp.zeros((8, 3)))

  def test_unbox(self):
    xs = {
      'kernel': meta.Partitioned(jnp.zeros((3, 2)), ('in', 'out')),
      'complex': meta.Partitioned(
        {'K': jnp.zeros((3, 2)), 'b': jnp.zeros((3,))}, ('data',)
      ),
    }
    unboxed = meta.unbox(xs)
    unboxed_shapes = jax.tree_util.tree_map(jnp.shape, unboxed)
    self.assertEqual(
      unboxed_shapes,
      {
        'kernel': (3, 2),
        'complex': {
          'K': (3, 2),
          'b': (3,),
        },
      },
    )

  def test_scan_over_layers(self):
    def f(scope, x):
      def body(scope, x):
        kernel_init = meta.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        y = nn.dense(scope, x, 3, kernel_init=kernel_init)
        return y, ()

      c, _ = lift.scan(
        body,
        variable_axes={'params': 0},
        split_rngs={'params': True},
        length=8,
        metadata_params={meta.PARTITION_NAME: 'layers'},
      )(scope, x)
      return c

    _, variables = init(f)(random.key(0), jnp.zeros((8, 3)))
    boxed_shapes = jax.tree_util.tree_map(jnp.shape, variables['params'])
    self.assertEqual(
      boxed_shapes,
      {
        'kernel': meta.Partitioned((8, 3, 3), ('layers', 'in', 'out')),
        'bias': (8, 3),
      },
    )

  def test_get_partition_spec(self):
    xs = {
      'kernel': meta.Partitioned(jnp.zeros((8, 3, 3)), ('layers', 'in', 'out')),
      'bias': jnp.zeros((8, 3)),
      'step': jnp.array(100),
    }
    ps = meta.get_partition_spec(xs)
    self.assertEqual(
      ps,
      {
        'kernel': jax.sharding.PartitionSpec('layers', 'in', 'out'),
        'bias': jax.sharding.PartitionSpec(),
        'step': jax.sharding.PartitionSpec(),
      },
    )

  def test_get_sharding(self):
    devices = mesh_utils.create_device_mesh((jax.local_device_count(), 1))
    mesh = sharding.Mesh(devices, ('in', 'out'))
    xs = {
      'kernel': meta.Partitioned(jnp.zeros((8, 3)), ('in', 'out')),
      'bias': jnp.zeros((8, 3)),
      'step': jnp.array(100),
    }
    ps = meta.get_sharding(xs, mesh)
    self.assertEqual(
      ps,
      {
        'kernel': jax.sharding.NamedSharding(
          mesh, jax.sharding.PartitionSpec('in', 'out')
        ),
        'bias': jax.sharding.NamedSharding(mesh, jax.sharding.PartitionSpec()),
        'step': jax.sharding.NamedSharding(mesh, jax.sharding.PartitionSpec()),
      },
    )

  def test_boxed_param_with_mesh(self):
    devices = mesh_utils.create_device_mesh((jax.local_device_count(), 1))
    mesh = sharding.Mesh(devices, ('in', 'out'))

    def f(scope, x):
      kernel_init = meta.with_partitioning(
        nn.initializers.ones_init(), ('in', 'out'), mesh=mesh
      )
      kernel = scope.param('kernel', kernel_init, (x.shape[-1], 2))
      kernel_box = scope.get_variable('params', 'kernel')
      self.assertIsInstance(kernel_box, meta.Partitioned)
      self.assertEqual(kernel_box.names, ('in', 'out'))
      return x @ kernel

    @jax.jit
    def create_state():
      y, variables = init(f)(random.key(0), jnp.zeros((8, 4)))
      spec = meta.get_partition_spec(variables)
      shardings = jax.tree_util.tree_map(
          lambda s: sharding.NamedSharding(mesh, s), spec
      )
      variables = jax.lax.with_sharding_constraint(variables, shardings)
      return variables

    variables = create_state()
    self.assertEqual(variables['params']['kernel'].names, ('in', 'out'))
    self.assertIs(variables['params']['kernel'].mesh, mesh)


if __name__ == '__main__':
  absltest.main()
